#ifndef _RMS_MESSAGE_CLIENT_H
#define _RMS_MESSAGE_CLIENT_H

#include <NovaRunControlClient/Action.h>
#include <NovaDAQUtilities/ConcurrentQueue.h>
#include <NovaDAQUtilities/Runnable.h>
#include <rms/RmsReceiver.h>
#include <rms/RmsSender.h>
#include <rms/provider/DDSConnection.h>
#include <DAQMessages/ccpp_RunControlMessages.h>
#include <NovaDAQUtilities/EnvVarCache.h>
#include <boost/shared_ptr.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/thread.hpp>

#include <vector>
#include <map>

// The meaning of the bits in the RMS_MESSAGE_CLIENT_DEBUG env var:
// - bit 0: reports when message handlers are installed

namespace RMSNS = gov::fnal::cd::rms;

namespace novadaq {

namespace rcclient {

// forward declaration of RmsMessageClient class
class RmsMessageClient;

// *******************************************************************
// Declaration of MessageNotifier class
// *******************************************************************

/**
 * The message notifier class is used by the RMS message client as
 * an encapsulation of the work that needs to get done when a listener
 * is notified that a message has been received.
 *
 * @author Kurt Biery
 * @version $Revision: 1.2 $ $Date: 2011/09/23 13:41:34 $
 */

template <class MSGTYPE, class LSTNR, class REPLYTYPE>
class MessageNotifier : public Action {

 public:
    MessageNotifier(RmsMessageClient* owner, int partitionNumber,
                    boost::shared_ptr<LSTNR> theListener,
                    boost::shared_ptr<MSGTYPE> theMessage);
    MessageNotifier(RmsMessageClient* owner, int partitionNumber,
                    LSTNR* theListener,
                    boost::shared_ptr<MSGTYPE> theMessage);
    void executeAction();
    int getPartitionNumber() {return _partitionNumber;}
    std::string getDescription() {
        std::ostringstream oss;
        oss << std::hex << ((uint64_t) this);

        std::string desc("MessageNotifier (0x");
        desc.append(oss.str());
        desc.append(") for ");
        desc.append(typeid(*_theMessage).name());
        return desc;
    }

 private:
    RmsMessageClient* _owner;
    int _partitionNumber;
    boost::shared_ptr<LSTNR> _listener;
    LSTNR* _listenerRawPtr;
    boost::shared_ptr<MSGTYPE> _theMessage;
};

// *******************************************************************
// Declaration of MessageReceiver class
// *******************************************************************

/**
 * The message receiver class is used by the RMS message client
 * to receive messages from RMS/DDS and put them on its internal queue
 * for later processing.
 *
 * @author Kurt Biery
 * @version $Revision: 1.2 $ $Date: 2011/09/23 13:41:34 $
 */

template <class MSGTYPE, class LSTNR, class REPLYTYPE>
class MessageReceiver {

 public:
    MessageReceiver(RmsMessageClient* owner, int partitionNumber,
                   boost::shared_ptr<LSTNR> theListener);
    MessageReceiver(RmsMessageClient* owner, int partitionNumber,
                   LSTNR* theListener);
    void processMessage(boost::shared_ptr<MSGTYPE> message);

 private:
    RmsMessageClient* _owner;
    int _partitionNumber;
    boost::shared_ptr<LSTNR> _listener;
    LSTNR* _listenerRawPtr;
    boost::mutex _callbackMutex;
};

// *******************************************************************
// Declaration of PartitionDisconnecter class
// *******************************************************************

/**
 * The partition disconnecter class is used by the RMS message client as
 * an encapsulation of the work that needs to get done when an application
 * disconnects from a partition.
 *
 * @author Kurt Biery
 * @version $Revision: 1.2 $ $Date: 2011/09/23 13:41:34 $
 */

class PartitionDisconnecter : public Action {

 public:
    PartitionDisconnecter(RmsMessageClient* owner, int partitionNumber);
    void executeAction();
    int getPartitionNumber() {return _partitionNumber;}
    std::string getDescription() {
        std::ostringstream oss;
        oss << std::hex << ((uint64_t) this);

        std::string desc("PartitionDisconnecter (0x");
        desc.append(oss.str());
        desc.append(")");
        return desc;
    }

 private:
    RmsMessageClient* _owner;
    int _partitionNumber;
};

// *******************************************************************
// Declaration of ReplySender class
// *******************************************************************

/**
 * The reply sender class is used by the RMS message client as
 * an encapsulation of the work that needs to get done when a reply
 * is sent to the requester.  Note that this is only used when application
 * code explicitly requests that a reply be sent.  For replies that are
 * handled in the normal callbacks when messages are received, the
 * sending of the reply is handled as part of processing the callback.
 *
 * @author Kurt Biery
 * @version $Revision: 1.2 $ $Date: 2011/09/23 13:41:34 $
 */

template <class REPLYTYPE>
class ReplySender : public Action {

 public:
    ReplySender(RmsMessageClient* owner, int partitionNumber,
                std::string replyTarget,
                boost::shared_ptr<REPLYTYPE> theReply);
    void executeAction();
    int getPartitionNumber() {return _partitionNumber;}
    std::string getDescription() {
        std::ostringstream oss;
        oss << std::hex << ((uint64_t) this);

        std::string desc("ReplySender (0x");
        desc.append(oss.str());
        desc.append(") for ");
        desc.append(typeid(*_theReply).name());
        return desc;
    }

 private:
    RmsMessageClient* _owner;
    int _partitionNumber;
    std::string _replyTarget;
    boost::shared_ptr<REPLYTYPE> _theReply;
};

// *******************************************************************
// Declaration of ReceiverWrapper class
// *******************************************************************

/**
 * The receiver wrapper class is used by the RMS message client to
 * help keep track of which message receivers are internally created
 * and which are created from externally-specified listeners.
 * (Internally-created ones get over-ridden by externally-specified ones,
 * but multiple externally-specified ones can co-exist.
 */

typedef boost::shared_ptr<RMSNS::base::RmsCloseable> CloseablePtr;

class ReceiverWrapper {

 public:
    ReceiverWrapper(CloseablePtr receiverPtr, const std::string& msgType,
                    const std::string& targetName, bool canBeOverridden);
    CloseablePtr getReceiverPtr() const {return _receiverPtr;}
    std::string getMsgType() const {return _msgType;}
    std::string getTargetName() const {return _targetName;}
    bool canBeOverridden() const {return _canBeOverridden;}

 private:
    CloseablePtr _receiverPtr;
    std::string _msgType;
    std::string _targetName;
    bool _canBeOverridden;
};

// *******************************************************************
// Declaration of DefaultListener class
// *******************************************************************

/**
 * The default listener class is used by the RMS message client as a
 * way to ensure that an application that uses the RMS Message Client will
 * always reply to all known messages for a given requester, even if the
 * application hasn't explicitly provide a handler for each message type.
 */

template <class MSGTYPE, class REPLYTYPE>
class DefaultListener {

 public:
    boost::shared_ptr<REPLYTYPE>
        handleMessage(boost::shared_ptr<MSGTYPE> message);
};

// *******************************************************************
// Declaration of RmsMessageClient class
// *******************************************************************

/**
 * The RMS message client handles the receipt of messages from an external
 * application and the sending of replies back to that application.  It does
 * this by allowing user code to specify callbacks that get invoked when
 * specified message types arrive.
 *
 * @author Kurt Biery
 * @version $Revision: 1.2 $ $Date: 2011/09/23 13:41:34 $
 */

class RmsMessageClient : public novadaq::Runnable {

 public:

    typedef boost::shared_ptr<RMSNS::provider::DDSConnection> DDSConnectionPtr;
    typedef std::map<int, DDSConnectionPtr> DDSConnectionMap;

    typedef std::vector<CloseablePtr> CloseableList;
    typedef boost::shared_ptr<CloseableList> CloseableListPtr;
    typedef std::map<int, CloseableListPtr> CloseableListMap;

    typedef boost::shared_ptr<ReceiverWrapper> ReceiverWrapperPtr;
    typedef std::vector<ReceiverWrapperPtr> ReceiverWrapperList;
    typedef boost::shared_ptr<ReceiverWrapperList> ReceiverWrapperListPtr;
    typedef std::map<int, ReceiverWrapperListPtr> ReceiverWrapperListMap;

    typedef std::map<std::string, CloseablePtr> CloseableMap;
    typedef boost::shared_ptr<CloseableMap> CloseableMapPtr;
    typedef std::map<int, CloseableMapPtr> CloseableMapMap;


    RmsMessageClient(const std::string& applicationName,
                     std::vector<std::string> bcastTargetList,
                     const std::string& rmsMessageChannel,
                     bool provideDefaultHandlers = false);

    virtual ~RmsMessageClient();

    template <class MSGTYPE, class LSTNR, class REPLYTYPE>
    void addListener(int partitionNumber, boost::shared_ptr<LSTNR> theListener);

    template <class MSGTYPE, class LSTNR, class REPLYTYPE>
    void addListener(int partitionNumber, LSTNR* theListener);

    void disconnectFromPartition(int partitionNumber);

    template <class REPLYTYPE>
    void sendAsyncReply(int partitionNumber, std::string replyTarget,
                        boost::shared_ptr<REPLYTYPE> theReply);

    void enqueueAction(boost::shared_ptr<Action> theAction);

    void run();

    template <class REPLYTYPE>
    void _sendReply(int partitionNumber, std::string replyTarget,
                    boost::shared_ptr<REPLYTYPE> theReply);

 private:

    template <class MSGTYPE, class LSTNR, class REPLYTYPE>
    void _addListener(int partitionNumber,
                     boost::shared_ptr<MessageReceiver<MSGTYPE, LSTNR,
                     REPLYTYPE> > localListener);

    template <class MSGTYPE, class LSTNR, class REPLYTYPE>
    void _addUserSpecifiedListener(DDSConnectionPtr ddsConnection,
                                   RMSNS::base::RmsDestination destination,
                                   ReceiverWrapperListPtr receiverListPtr,
                                   boost::shared_ptr<MessageReceiver<MSGTYPE,
                                   LSTNR, REPLYTYPE> > listener);

 protected:

    virtual bool _thisInstanceShouldInstallDefaultHandlers() = 0;

    virtual void
        _addDefaultListeners(DDSConnectionPtr ddsConnection,
                             RMSNS::base::RmsDestination destination,
                             int partitionNumber,
                             ReceiverWrapperListPtr receiverListPtr) = 0;

    template <class MSGTYPE, class REPLYTYPE>
    void _addOneDefaultListener(DDSConnectionPtr ddsConnection,
                                RMSNS::base::RmsDestination destination,
                                int partitionNumber,
                                ReceiverWrapperListPtr receiverListPtr);

 private:

    void _disconnectFromPartition(int partitionNumber);

    template <class REPLYTYPE>
    CloseablePtr _getReplyRmsSender(int partitionNumber,
                                    std::string replyTarget,
                                    boost::shared_ptr<REPLYTYPE> theReply);

    // the application name
    std::string _applicationName;

    // list of broadcast target names that we should listen to
    std::vector<std::string> _bcastTargetList;

    // the RMS "channel" to use when sending/receiving messages
    std::string _rmsMessageChannel;

    // flag that indicates whether we should try to install default handlers
    bool _provideDefaultHandlers;

    // map of partition numbers to DDSConnections
    DDSConnectionMap _connectionTable;

    // map of partition numbers to lists of RmsReceivers
    ReceiverWrapperListMap _receiverListTable;

    // map of partition numbers to lists of RmsSenders
    CloseableMapMap _senderListTable;

    // mutex for protecting access to lists of connections, senders, receivers
    boost::mutex _workListMutex;

    // queue of Actions that is used to manage the work that needs to be done
    ConcurrentQueue< boost::shared_ptr<Action> > _workQueue;

    // the thread that does the work for the receiver
    boost::thread* _workerThread;

    int _debugFlag;

    friend class PartitionDisconnecter;
};

// *******************************************************************
// Definition of MessageNotifier class
// *******************************************************************

/**
 * Creates a MessageNotifier instance from the specified parameters,
 * including a shared pointer to the listener that should be invoked
 * when a message arrives.
 */
template <class MSGTYPE, class LSTNR, class REPLYTYPE>
MessageNotifier<MSGTYPE, LSTNR, REPLYTYPE>::
MessageNotifier(RmsMessageClient* owner, int partitionNumber,
                boost::shared_ptr<LSTNR> theListener,
                boost::shared_ptr<MSGTYPE> theMessage)
{
    _owner = owner;
    _partitionNumber = partitionNumber;
    _listener = theListener;
    _listenerRawPtr = 0;
    _theMessage = theMessage;
}

/**
 * Creates a MessageNotifier instance from the specified parameters,
 * including a bare pointer to the listener that should be invoked
 * when a message arrives.
 */
template <class MSGTYPE, class LSTNR, class REPLYTYPE>
MessageNotifier<MSGTYPE, LSTNR, REPLYTYPE>::
MessageNotifier(RmsMessageClient* owner, int partitionNumber,
                LSTNR* theListener,
                boost::shared_ptr<MSGTYPE> theMessage)
{
    _owner = owner;
    _partitionNumber = partitionNumber;
    _listenerRawPtr = theListener;
    _theMessage = theMessage;
}

/**
 * Does the work asosciated with notifying a listener that a
 * message has arrived.
 */
template <class MSGTYPE, class LSTNR, class REPLYTYPE>
void MessageNotifier<MSGTYPE, LSTNR, REPLYTYPE>::executeAction()
{
    //std::cout << "MessageNotifier::executeAction " << _partitionNumber
    //          << " " << "MSGTYPE" << std::endl;
    boost::shared_ptr<REPLYTYPE> replyMessage;
    if (_listener.get() != 0) {
        replyMessage = _listener->handleMessage(_theMessage);
    }
    else if (_listenerRawPtr != 0) {
        replyMessage = _listenerRawPtr->handleMessage(_theMessage);
    }

    if (replyMessage.get() != 0) {
        gov::fnal::cd::rms::
            fillCorrelationFields<MSGTYPE, REPLYTYPE>(*_theMessage,
                                                      *replyMessage);

        std::string replyTarget(_theMessage->header.replyTarget);

        _owner->_sendReply(_partitionNumber, replyTarget, replyMessage);
    }
}

// *******************************************************************
// Definition of MessageReceiver class
// *******************************************************************

/**
 * Creates a MessageReceiver instance from the specified parameters,
 * including a shared pointer to the listener that should be invoked
 * when a message arrives.
 */
template <class MSGTYPE, class LSTNR, class REPLYTYPE>
MessageReceiver<MSGTYPE, LSTNR, REPLYTYPE>::
MessageReceiver(RmsMessageClient* owner, int partitionNumber,
               boost::shared_ptr<LSTNR> theListener)
{
    _owner = owner;
    _partitionNumber = partitionNumber;
    _listener = theListener;
    _listenerRawPtr = 0;
}

/**
 * Creates a MessageReceiver instance from the specified parameters,
 * including a bare pointer to the listener that should be invoked
 * when a message arrives.
 */
template <class MSGTYPE, class LSTNR, class REPLYTYPE>
MessageReceiver<MSGTYPE, LSTNR, REPLYTYPE>::
MessageReceiver(RmsMessageClient* owner, int partitionNumber,
               LSTNR* theListener)
{
    _owner = owner;
    _partitionNumber = partitionNumber;
    _listenerRawPtr = theListener;
}

/**
 * Receives a message from the RMS/DDS system and puts it on the
 * RMS message client work queue for later processing.
 */
template <class MSGTYPE, class LSTNR, class REPLYTYPE>
void MessageReceiver<MSGTYPE, LSTNR, REPLYTYPE>::
processMessage(boost::shared_ptr<MSGTYPE> message)
{
    boost::mutex::scoped_lock sl(_callbackMutex);

    boost::shared_ptr< MessageNotifier<MSGTYPE, LSTNR, REPLYTYPE> > notifier;

    //std::cout << "MessageReceiver::executeAction " << _partitionNumber
    //          << " " << "MSGTYPE" << std::endl;
    if (_listener.get() != 0) {
        notifier.reset(new MessageNotifier<MSGTYPE, LSTNR,
                       REPLYTYPE>(_owner, _partitionNumber,
                                  _listener, message));
    }
    else if (_listenerRawPtr != 0) {
        notifier.reset(new MessageNotifier<MSGTYPE, LSTNR,
                       REPLYTYPE>(_owner, _partitionNumber,
                                  _listenerRawPtr, message));
    }

    if (notifier.get() != 0) {
        _owner->enqueueAction(notifier);
    }
}

// *******************************************************************
// Definition of PartitionDisconnecter class
// *******************************************************************

/**
 * Creates a PartitionDisconnector instance for the specified partition.
 */
inline PartitionDisconnecter::
PartitionDisconnecter(RmsMessageClient* owner, int partitionNumber)
{
    _owner = owner;
    _partitionNumber = partitionNumber;
}

/**
 * Does the work of actually disconnection the RMS message client
 * from a partition.
 */
inline void PartitionDisconnecter::executeAction()
{
    _owner->_disconnectFromPartition(_partitionNumber);
}

// *******************************************************************
// Definition of ReplySender class
// *******************************************************************

/**
 * Creates a ReplySender instance with the specified parameters.
 */
template <class REPLYTYPE>
ReplySender<REPLYTYPE>::
ReplySender(RmsMessageClient* owner, int partitionNumber,
            std::string replyTarget,
            boost::shared_ptr<REPLYTYPE> theReply)
{
    _owner = owner;
    _partitionNumber = partitionNumber;
    _replyTarget = replyTarget;
    _theReply = theReply;
}

/**
 * Does the work of sending a reply.
 */
template <class REPLYTYPE> void ReplySender<REPLYTYPE>::executeAction()
{
    _owner->_sendReply(_partitionNumber, _replyTarget, _theReply);
}

// *******************************************************************
// Definition of DefaultListener class
// *******************************************************************

/**
 * Creates a default reply to send back to the requester.
 */
template <class MSGTYPE, class REPLYTYPE>
boost::shared_ptr<REPLYTYPE> DefaultListener<MSGTYPE, REPLYTYPE>::
handleMessage(boost::shared_ptr<MSGTYPE> message)
{
    // initialize the status that we'll send in the reply to sucess
    rmscore::RequestStatus reqStatus;
    reqStatus.code = 0;
    reqStatus.message = DDS::string_dup("Success (default handler)");

    // create the reply
    boost::shared_ptr<REPLYTYPE> transResponse(new REPLYTYPE());
    transResponse->status = reqStatus;

    return transResponse;
}

// *******************************************************************
// Definition of RmsMessageClient class
// *******************************************************************

/**
 * Adds the specified callback to the receiver for the
 * specified partition number.
 */
template <class MSGTYPE, class LSTNR, class REPLYTYPE>
void RmsMessageClient::addListener(int partitionNumber,
                                   boost::shared_ptr<LSTNR> theListener) {
    // create the local listener
    boost::shared_ptr<MessageReceiver<MSGTYPE, LSTNR, REPLYTYPE> >
        localListener(new MessageReceiver<MSGTYPE, LSTNR, REPLYTYPE>
                      (this, partitionNumber, theListener));

    // call our private addListener method to do the real work
    this->_addListener(partitionNumber, localListener);
}

/**
 * Adds the specified callback to the receiver for the
 * specified partition number.
 */
template <class MSGTYPE, class LSTNR, class REPLYTYPE>
void RmsMessageClient::addListener(int partitionNumber,
                                   LSTNR* theListener) {
    // create the local listener
    boost::shared_ptr<MessageReceiver<MSGTYPE, LSTNR, REPLYTYPE> >
        localListener(new MessageReceiver<MSGTYPE, LSTNR, REPLYTYPE>
                      (this, partitionNumber, theListener));

    // call our private addListener method to do the real work
    this->_addListener(partitionNumber, localListener);
}

/**
 * Sends the specified asynchronous reply to the requester for the
 * specified partition number.
 */
template <class REPLYTYPE>
void RmsMessageClient::sendAsyncReply(int partitionNumber,
                                      std::string replyTarget,
                                      boost::shared_ptr<REPLYTYPE> theReply) {

    // create the request to send the reply
    boost::shared_ptr<Action>
        replyRequest(new ReplySender<REPLYTYPE>(this, partitionNumber,
                                                replyTarget, theReply));

    // enqueue the request
    this->enqueueAction(replyRequest);
}

/**
 * Adds the specified listener to this receiver instance for
 * the specified partition number.
 */
template <class MSGTYPE, class LSTNR, class REPLYTYPE>
void RmsMessageClient::
_addListener(int partitionNumber,
             boost::shared_ptr< MessageReceiver<MSGTYPE, LSTNR,
             REPLYTYPE> > localListener) {
    boost::mutex::scoped_lock sl(_workListMutex);

    // create the DDS connection, if needed
    DDSConnectionPtr ddsConnection;
    DDSConnectionMap::iterator connIter =
        _connectionTable.find(partitionNumber);
    if (connIter == _connectionTable.end()) {
        ddsConnection.reset(new RMSNS::provider::
                            DDSConnection(_applicationName,
                                          partitionNumber));
        _connectionTable[partitionNumber] = ddsConnection;
    }
    else {
        ddsConnection = connIter->second;
    }

    // create the initial, empty list of receivers, if needed
    bool defaultHandlersArePossible = false;
    ReceiverWrapperListPtr receiverListPtr;
    ReceiverWrapperListMap::iterator recIter =
        _receiverListTable.find(partitionNumber);
    if (recIter == _receiverListTable.end()) {
        receiverListPtr.reset(new ReceiverWrapperList());
        _receiverListTable[partitionNumber] = receiverListPtr;

        if (_thisInstanceShouldInstallDefaultHandlers()) {
            defaultHandlersArePossible = true;
        }
    }
    else {
        receiverListPtr = recIter->second;
    }

    // subscribe to messages sent directly to us
    RMSNS::base::RmsDestination directMsgDest(_applicationName,
                                              _rmsMessageChannel);
    directMsgDest.setPartitionProperty(partitionNumber);

    _addUserSpecifiedListener(ddsConnection, directMsgDest,
                              receiverListPtr, localListener);

    if ((_debugFlag & 0x01) != 0) {
        std::cout << "*** RmsMessageClient(0x" << std::hex
                  << this << std::dec << "): "
                  << "provideDefaultHandlers = " << _provideDefaultHandlers
                  << ", defaultHandlersArePossible = "
                  << defaultHandlersArePossible
                  << ", partition = " << partitionNumber << std::endl;
    }
    if (_provideDefaultHandlers && defaultHandlersArePossible &&
        partitionNumber >= 0) {
        _addDefaultListeners(ddsConnection, directMsgDest,
                             partitionNumber, receiverListPtr);
    }

    // subscribe to messages sent to broadcast targets that we care about
    std::vector<std::string>::iterator bcastIter;
    for (bcastIter = _bcastTargetList.begin();
         bcastIter != _bcastTargetList.end();
         bcastIter++) {
        std::string bcastTarget = *bcastIter;

        RMSNS::base::RmsDestination bcastDest(bcastTarget, _rmsMessageChannel);
        bcastDest.setPartitionProperty(partitionNumber);

        _addUserSpecifiedListener(ddsConnection, bcastDest,
                                  receiverListPtr, localListener);

        if (_provideDefaultHandlers && defaultHandlersArePossible &&
            partitionNumber >= 0) {
            _addDefaultListeners(ddsConnection, bcastDest,
                                 partitionNumber, receiverListPtr);
        }
    }
}

/**
 * Adds a user-specified listener to this receiver instance for
 * the specified destination, taking care to handle over-rides
 * and duplicates correctly.
 */
template <class MSGTYPE, class LSTNR, class REPLYTYPE>
void RmsMessageClient::
_addUserSpecifiedListener(DDSConnectionPtr ddsConnection,
                          RMSNS::base::RmsDestination destination,
                          ReceiverWrapperListPtr receiverListPtr,
                          boost::shared_ptr<MessageReceiver<MSGTYPE,
                          LSTNR, REPLYTYPE> > listener) {

    boost::shared_ptr< RMSNS::RmsReceiver<RMSNS::provider::DDSConnection,
        MSGTYPE, MessageReceiver<MSGTYPE, LSTNR, REPLYTYPE> > > rmsReceiver;
    rmsReceiver.reset(new RMSNS::RmsReceiver<RMSNS::provider::
                      DDSConnection, MSGTYPE, 
                      MessageReceiver<MSGTYPE, LSTNR, REPLYTYPE> >
                      (ddsConnection, destination));
    rmsReceiver->setListener(listener);

    MSGTYPE* dummyMsg;
    std::string msgType = typeid(*dummyMsg).name();
    std::string targetName =
        destination.getProperty(RMSNS::base::RmsDestination::
                                TARGET_PROPERTY_NAME, "unknown");
    std::string partNumString =
        destination.getProperty(RMSNS::base::RmsDestination::
                                PARTITION_PROPERTY_NAME, "unknown");
    ReceiverWrapperPtr
        receiverWrapperPtr(new ReceiverWrapper(rmsReceiver, msgType,
                                               targetName, false));
    receiverListPtr->push_back(receiverWrapperPtr);
    if ((_debugFlag & 0x01) != 0) {
        std::cout << "*** RmsMessageClient(0x" << std::hex
                  << this << std::dec << "): "
                  << "Installed user-specified handler for "
                  << msgType << ", " << targetName
                  << ", partition " << partNumString << std::endl;
    }

    // remove any existing over-rideable receivers with the same
    // message type and target name as this one
    for (int idx = receiverListPtr->size() - 1; idx >= 0; --idx) {
        ReceiverWrapperPtr tmpWrapperPtr = receiverListPtr->at(idx);
        if (tmpWrapperPtr->canBeOverridden() &&
            tmpWrapperPtr->getMsgType() == receiverWrapperPtr->getMsgType() &&
            tmpWrapperPtr->getTargetName() ==
            receiverWrapperPtr->getTargetName()) {
            receiverListPtr->erase(receiverListPtr->begin() + idx);
            if ((_debugFlag & 0x01) != 0) {
                std::cout << "*** RmsMessageClient(0x" << std::hex
                          << this << std::dec << "): "
                          << "Dropped default handler for "
                          << msgType << ", " << targetName
                          << ", partition " << partNumString << std::endl;
            }
        }
    }
}

/**
 * Adds a default listener to this receiver instance for
 * the specified destination, taking care to handle over-rides
 * and duplicates correctly.
 */
template <class MSGTYPE, class REPLYTYPE>
void RmsMessageClient::
_addOneDefaultListener(DDSConnectionPtr ddsConnection,
                       RMSNS::base::RmsDestination destination,
                       int partitionNumber,
                       ReceiverWrapperListPtr receiverListPtr) {

    // only add a default handler if there isn't already a handler
    // for the same msgType and targetName in the list
    MSGTYPE* dummyMsg;
    std::string msgType = typeid(*dummyMsg).name();
    std::string targetName =
        destination.getProperty(RMSNS::base::RmsDestination::
                                TARGET_PROPERTY_NAME, "unknown");
    for (int idx = 0; idx < (int) receiverListPtr->size(); ++idx) {
        ReceiverWrapperPtr tmpWrapperPtr = receiverListPtr->at(idx);
        if (tmpWrapperPtr->getMsgType() == msgType &&
            tmpWrapperPtr->getTargetName() == targetName) {
            if ((_debugFlag & 0x01) != 0) {
                std::cout << "*** RmsMessageClient(0x" << std::hex
                          << this << std::dec << "): "
                          << "Skipped install of default handler for "
                          << msgType << ", " << targetName
                          << ", partition " << partitionNumber << std::endl;
            }
            return;
        }
    }

    boost::shared_ptr< DefaultListener<MSGTYPE, REPLYTYPE> >
        defaultListener(new DefaultListener<MSGTYPE, REPLYTYPE>());

    boost::shared_ptr< MessageReceiver<MSGTYPE,
        DefaultListener<MSGTYPE, REPLYTYPE>, REPLYTYPE> >
        messageReceiver(new MessageReceiver<MSGTYPE,
                    DefaultListener<MSGTYPE, REPLYTYPE>, REPLYTYPE>
                    (this, partitionNumber, defaultListener));

    boost::shared_ptr< RMSNS::RmsReceiver<RMSNS::provider::DDSConnection,
        MSGTYPE, MessageReceiver<MSGTYPE, DefaultListener<MSGTYPE, REPLYTYPE>,
        REPLYTYPE> > > rmsReceiver;
    rmsReceiver.reset(new RMSNS::RmsReceiver<RMSNS::provider::
                      DDSConnection, MSGTYPE, MessageReceiver<MSGTYPE,
                      DefaultListener<MSGTYPE, REPLYTYPE>, REPLYTYPE> >
                      (ddsConnection, destination));
    rmsReceiver->setListener(messageReceiver);

    ReceiverWrapperPtr
        receiverWrapperPtr(new ReceiverWrapper(rmsReceiver, msgType,
                                               targetName, true));
    receiverListPtr->push_back(receiverWrapperPtr);
    if ((_debugFlag & 0x01) != 0) {
        std::cout << "*** RmsMessageClient(0x" << std::hex
                  << this << std::dec << "): "
                  << "Installed default handler for "
                  << msgType << ", " << targetName
                  << ", partition " << partitionNumber << std::endl;
    }
}

/**
 * Sends the specified reply message to the specified target and partition.
 */
template <class REPLYTYPE>
void RmsMessageClient::_sendReply(int partitionNumber,
                                  std::string replyTarget,
                                  boost::shared_ptr<REPLYTYPE> theReply) {
    CloseablePtr senderPtr =
        this->_getReplyRmsSender<REPLYTYPE>(partitionNumber,
                                            replyTarget, theReply);
    if (senderPtr.get() != 0) {
        RMSNS::RmsSender<RMSNS::provider::DDSConnection, REPLYTYPE>*
            replySender = dynamic_cast< RMSNS::RmsSender<RMSNS::provider::
            DDSConnection, REPLYTYPE>* >(senderPtr.get());
        replySender->sendMessage(*theReply);
    }
    //else {
    //    std::cout << "Skipping reply because there is no sender" << std::endl;
    //    std::cout.flush();
    //}
}

/**
 * Looks up the reply sender instance that should be used for the
 * specified partition number and target.
 */
template <class REPLYTYPE>
CloseablePtr RmsMessageClient::_getReplyRmsSender(int partitionNumber,
                               std::string replyTarget,
                               boost::shared_ptr<REPLYTYPE> theReply) {
    boost::mutex::scoped_lock sl(_workListMutex);
    CloseablePtr senderPtr;

    // look up the DDS connection
    DDSConnectionPtr ddsConnection;
    DDSConnectionMap::iterator connIter =
        _connectionTable.find(partitionNumber);
    if (connIter == _connectionTable.end()) {
        return senderPtr;
    }
    ddsConnection = connIter->second;

    // find the list of senders for the partition, create it if necessary
    CloseableMapPtr senderMapPtr;
    CloseableMapMap::iterator mapIter = _senderListTable.find(partitionNumber);
    if (mapIter == _senderListTable.end()) {
        senderMapPtr.reset(new CloseableMap());
        _senderListTable[partitionNumber] = senderMapPtr;
    }
    else {
        senderMapPtr = mapIter->second;
    }

    // find the desired sender, create it if necessary
    std::string key(replyTarget);
    key.append(typeid(*theReply).name());
    CloseableMap::iterator sndrIter = senderMapPtr->find(key);
    if (sndrIter == senderMapPtr->end()) {
        RMSNS::base::RmsDestination dest(replyTarget, _rmsMessageChannel);

        senderPtr.reset(new RMSNS::RmsSender<RMSNS::provider::DDSConnection,
                        REPLYTYPE>(ddsConnection, dest));
        senderMapPtr->insert(make_pair(key, senderPtr));
    }
    else {
        senderPtr = sndrIter->second;
    }

    return senderPtr;
}

};  // end of namespace rcclient

};  // end of namespace novadaq

#endif
