/*
 * SNEWSMessageInfo header file 
 */
#ifndef __SNEWS_MESSAGEINFO_H_
#define __SNEWS_MESSAGEINFO_H_

#include <list>
#include <stdint.h>
#include <string>

/**
 * Class SNEWSMessageInfo
 *
 * This class holds all information for a message
 * time, type, and location.
 *
 * Times are stored in Nova time, 64 MHz ticks since nova epoch.
 * GPS time is different from UTC as explained in
 * http://tycho.usno.navy.mil/gpstt.html and elsewhere.
 * Note that GPS time, TAI (International Atomic time) and UTC all differ.
 * 
 * Original:
 * @author H. Meyer
 * @author <a href="mailto:holger.meyer@wichita.edu">holger.meyer@wichita.edu</a>
 *
 * Adapted:
 * J. Zirnstein (janzirn@fnal.gov) on 2014-04-01
 */

namespace nova {
namespace NovaSNEWSInterface {

enum MessageType{
  kSNEWSTrigger,	// Actual SNEWS Trigger
  kSNEWSBeatSlow,	// Heart Beat once per day
  kSNEWSBeatFast,	// Heart Beat once per minute
  kTestConnection,	// Testing the connection
  kFake,		// Will find a use for this (maybe)
  kNMessageType		// Number of Message types
};


class SNEWSMessageInfo {

 public:

  SNEWSMessageInfo () {;}
  SNEWSMessageInfo (uint64_t const time, int const type);

  static int getMessageTypeFromName(std::string const& name);
  static std::string getMessageNameFromType(int type);
  static std::list< int > allMessageTypes();
  
  void set(uint64_t const time, int const type);

  uint64_t getTime() const {return _messagetime;}
  enum MessageType getType() const {return _messagetype;}


 private:

  enum MessageType _messagetype;
  uint64_t _messagetime; // see the NovaTimingUtilities package

  bool _checkMessageTypeRange(int type);

};

} // end of namespace NovaSNEWSInterface
} // end of namespace nova

#endif // __SNEWS_MESSAGEINFO_H_
