/* A thread dealing with Data Driven Triggers */
/* Triggers come from Buffer Nodes, there are different */
/* types of DataTriggers to watch for from the BNs. */

// STD Libs
#include <string>
#include <map>
#include <deque>

// Boost Libs
#include <boost/shared_ptr.hpp>
#include <boost/assign.hpp>

// Class Def
#include <NovaGlobalTrigger/GTDataTrig.h>


// Nova Message Facility Libs
#include <messagefacility/MessageLogger/MessageLogger.h>

// DAQ Data Formats Lib
#include <DAQDataFormats/RawTriggerHeader.h>
#include <DAQDataFormats/RawTriggerTime.h>
#include <DAQDataFormats/RawTriggerTimingMarker.h>
#include <DAQDataFormats/RawTriggerRange.h>
#include <DAQDataFormats/RawTriggerMask.h>
#include <DAQDataFormats/RawTrigger.h>
#include <DAQDataFormats/TriggerDefines.h>

// Nova Utilities Libs
#include <NovaTimingUtilities/TimingUtilities.h>

// Inter process Messages
#include <DAQMessages/ccpp_GTMessages.h>
#include <DAQMessages/ccpp_DDTMessages.h>

// Global Trigger Defines
#include <NovaGlobalTrigger/GTGlobals.h>
#include <NovaGlobalTrigger/GT_Stats.h>
#include <NovaGlobalTrigger/GTLongTrigs.h>
#include <NovaGlobalTrigger/GT_SendQueues.h>

// RMS Libs
#include <rms/provider/DDSConnection.h>
#include <rms/base/RmsDestination.h>
#include <rms/RmsReceiver.h>

using namespace std;
using namespace gov::fnal::cd::rms;
using namespace novadaq;
using namespace gt;
using namespace daqdataformats;

#define TRIGGER_APP_NAME "GTDataTrig"

// Helper Function for Logging Trigger Setup Info
bool GTDataTrig::LogEnabledTrigs(trigID trig_label){
  if (GT_Stats::getInstance().Triggers.at(trig_label).getEnabled()) {
    ::mf::LogInfo("Progress") << "Data Driven Trigger named "
			      << GT_Stats::getInstance().Triggers.at(trig_label).getName()
			      << " enabled with prescale " 
			      << GT_Stats::getInstance().Triggers.at(trig_label).getPrescale()
			      << " queue depth " 
			      << GT_Stats::getInstance().Triggers.at(trig_label).Queue->getMaxQueueDepth()
			      << " max rate "
			      << GT_Stats::getInstance().Triggers.at(trig_label).Queue->getMaxRate()
			      << endl;      
    return true;
  }
  return false;
}

void GTDataTrig::configure(){

  setname(std::string("data driven triggers"));

  // Loop over the trigger names and figure out all the defined ones
  for(int i=0; i < 88; ++i){
    
  }

  // Mapping of Strings to IDs in the trigger
  setTrigNameMap(
    boost::assign::map_list_of<std::string,daqdataformats::trigID>
    ("DDSN"       ,TRIG_ID_DATA_SN)
    ("DDenergy"   ,TRIG_ID_DATA_ENERGY)
    ("DDcalmu"    ,TRIG_ID_DATA_CALMU)
    ("DDupmu"     ,TRIG_ID_DATA_UPMU)
    ("DDnumu"     ,TRIG_ID_DATA_NUMU)
    ("DDnue"      ,TRIG_ID_DATA_NUE)
    ("DDfastmono" ,TRIG_ID_DATA_FASTMONO)
    ("DDslowmono" ,TRIG_ID_DATA_SLOWMONO)
    ("FEBOverflow",TRIG_ID_DATA_FEBOVERFLOW)
    ("DDActivity1",TRIG_ID_DATA_ACTIVITY_1)
    ("DDActivity2",TRIG_ID_DATA_ACTIVITY_2)
    ("DDContained",TRIG_ID_DATA_CONTAINED)
    ("DDMoon",     TRIG_ID_DATA_MOON)
    ("DDSun",      TRIG_ID_DATA_SUN)
  );
  // Mapping of Strings to IDs in the trigger
  setValidTrigTypes(
    boost::assign::map_list_of<daqdataformats::trigID,bool>
    (TRIG_ID_DATA_SN,true)
    (TRIG_ID_DATA_ENERGY,true)
    (TRIG_ID_DATA_CALMU,true)
    (TRIG_ID_DATA_UPMU,true)
    (TRIG_ID_DATA_NUMU,true)
    (TRIG_ID_DATA_NUE,true)
    (TRIG_ID_DATA_FASTMONO,true)
    (TRIG_ID_DATA_SLOWMONO,true)
    (TRIG_ID_DATA_FEBOVERFLOW,true)
    (TRIG_ID_DATA_ACTIVITY_1,true)
    (TRIG_ID_DATA_ACTIVITY_2,true)
    (TRIG_ID_DATA_CONTAINED,true)
    (TRIG_ID_DATA_MOON,true)
    (TRIG_ID_DATA_SUN,true)
  );
  
  return;
}

void GTDataTrig::work() {

    mf::SetModuleName("GTDataTrig");

    // Get the instance of the sending queues
    GT_SendQueues& theQueues = gt::GT_SendQueues::getInstance();

    {
      boost::mutex::scoped_lock lock(gtCfgSet_mutex);

      bool somethingIsEnabled = false;
 
      ValidTrigTypes.at(TRIG_ID_DATA_SN);

      // Grab the trigger config from global memory
      // This has to be there, at least one trigger is mandatory
      const TriggerType trigCfg = gtCfgSet.getGTConfiguration().getTriggers();
      // Loop over list of Data Driven Triggers
      if (trigCfg.hasDDTSet()) {
	if (trigCfg.getDDTSet().hasDDT()) {
	  int ddtCfgCount = trigCfg.getDDTSet().getDDTCount();	
	  LOG_DEBUG("Progress") << "There are " << ddtCfgCount
				<< " DDTs listed" << endl;
	  for (int idx = 0; idx < ddtCfgCount; ++idx) {
	    DDTType ddtCfg = trigCfg.getDDTSet().getDDT(idx);

	    // parse name into trigger number
	    trigID trigType = TRIG_ID_NULL;
	    std::string ddtName = ddtCfg.getName();

	    try{
	      trigType = TrigNameMap.at(ddtName);
	    }catch(...){
	      LOG_ERROR("DDTConfig") << ddtName << " is an unknown DDT" << endl;
	    }

	    // setup trigger
	    if (trigType != TRIG_ID_NULL) {
	      GTInfo *info = &GT_Stats::getInstance().Triggers.at(trigType);
	      info->setDDT(true);
	      info->setEnabled(ddtCfg.getEnableFlag());
	      if (info->getEnabled()){
		somethingIsEnabled = true;
	      } // endif
	      info->setPrescale(ddtCfg.getPrescale());
	      info->setName(ddtCfg.getName());
	      info->setType(trigType);
	      if (info->getEnabled()){
		info->initMetric();
		info->Queue = new throttlePipeline;
		info->Queue->setMaxQueueDepth(ddtCfg.getDepth());
		info->Queue->setMaxRate(ddtCfg.getMaxrate());
	      } // endif
	    } // endif trigType
	  } // endfor idx
	} // endif getDDTSet
      } // endif trigCfg.hasDDTSet

      // Bail if no DataDriven Triggers are enabled
      if (!somethingIsEnabled) {
	return;
      }

      // Leave the configuration scope and release the mutex
    }  // end of mutexed config access


    // Log which triggers are enabled
    LogEnabledTrigs(TRIG_ID_DATA_SN);
    LogEnabledTrigs(TRIG_ID_DATA_ENERGY);
    LogEnabledTrigs(TRIG_ID_DATA_CALMU);
    LogEnabledTrigs(TRIG_ID_DATA_UPMU);
    LogEnabledTrigs(TRIG_ID_DATA_NUMU);
    LogEnabledTrigs(TRIG_ID_DATA_NUE);
    LogEnabledTrigs(TRIG_ID_DATA_FASTMONO);
    LogEnabledTrigs(TRIG_ID_DATA_SLOWMONO);
    LogEnabledTrigs(TRIG_ID_DATA_FEBOVERFLOW);
    LogEnabledTrigs(TRIG_ID_DATA_ACTIVITY_1);
    LogEnabledTrigs(TRIG_ID_DATA_ACTIVITY_2);
    LogEnabledTrigs(TRIG_ID_DATA_CONTAINED);
    LogEnabledTrigs(TRIG_ID_DATA_MOON);
    LogEnabledTrigs(TRIG_ID_DATA_SUN);

    // create the RMS connection for the DataDriven processes
    // Must use the same partition number!
    boost::shared_ptr<provider::DDSConnection>
        rmsDDTConnection(new provider::DDSConnection("GTDDTListener",
                                    partNumber));
    // create the destination that represents where we will receive
    // messages from

    base::RmsDestination receiveDest("DDTServerMailbox",
                                     base::RmsDestination::TRIGGER_CHANNEL);
    // create a receiver with the connection object and destination
    RmsReceiver<provider::DDSConnection, ddtmessages::DDTTriggerMessage>
        requestReceiver(rmsDDTConnection, receiveDest);


    // loop over DataDriven triggers
    while (true) {
      // If we're not triggering, wait
      while (!Triggering) {
        boost::mutex::scoped_lock waitlock(Wait_mutex);
        Triggering_cond.wait(waitlock);
      }

      ddtmessages::DDTTriggerMessage ddtMessage;
      
      //grab msgs and store in ddtMessage
      requestReceiver.receiveMessage(ddtMessage);

      // If that trigger isn't enabled at the source skip to next one
      if (!ddtMessage.enabled) {
	continue;
      }

      // find out which sort of trigger it is.  The case setup here is
      // to make sure bogus types get caught, so we don't point to a 
      // bogus index in the Triggers array
      // question for Andrew: does this new control structure bail on a 
      // "continue"-like way if it catches a bogus trigger?
      trigID trigType = TRIG_ID_NULL; 

      try{
	// Validate that the trigger type that was sent matches our list
	if(ValidTrigTypes.at((daqdataformats::trigID)ddtMessage.trigger_type)){
	  trigType = (daqdataformats::trigID) ddtMessage.trigger_type;
	}else{
	  // The trigger type is marked invalid
	  trigType = TRIG_ID_NULL;
	  ::mf::LogWarning("Invalid Trigger Type") << "A DDT process has sent a trigger for an invalid trigger ID: " 
						   << ddtMessage.trigger_type;
	}
      }catch(...){
         ::mf::LogWarning("Unknown Trigger Type") << "A DDT process has sent an unknown trigger type, ID " 
						  << ddtMessage.trigger_type;
	trigType = TRIG_ID_NULL;
	 // Add buffer node ID and pid to this?
      }

      /***********************************************************/  
      /******************** Exception ****************************/  
      /***********************************************************/  

      GTInfo *info;
      try{
	info = &GT_Stats::getInstance().Triggers.at(trigType);
      }catch(...){

	::mf::LogWarning("Unknown Trigger Type") << "Attempted to get Stats for trigType: " 
						 << trigType;
	 // Add buffer node ID and pid to this?
      }

      // If that trigger isn't enabled in the config skip to next one
      if (!(info->getEnabled())){ continue; }

      if (!(info->incSeenCount() % info->getPrescale())) {
	// Get the time out, convert back to t_time
	uint64_t ddtNova = ddtMessage.start_time;

	timeval ddtTv;
	novadaq::timeutils::convertNovaTimeToUnixTime(ddtNova,ddtTv);
	time_t trigTime = ddtTv.tv_sec;

	// Determine when this trigger came in (to the GT)
	timeval trigArrival;
	gettimeofday(&trigArrival,0);

	// Calculation of rolling trigger rates
	// Push this time value into the overall DDT throttle
	DDTTrigThrottle.pushFront(trigArrival);
	
	// Push this time value into the trigger specific throttle
	info->Queue->pushFront(trigArrival);
	
	// Check to see if our average trigger rate has spiked above our maximum rate
	if(DDTTrigThrottle.throttleState() 
	  || info->Queue->throttleState()){
  	  // We have execeeded our throttle (either globally or on a given trigger
	  // Issue a warning message
	  if (DDTTrigThrottle.throttleState())
	    ::mf::LogWarning("Trigger Rate") << "Overall DDT Rate Average: " << DDTTrigThrottle.getRate() 
					     << "\nOverall Driven Trigger rate exceeds maximum, throttling trigger output.\n" 
					     << "Type of last trigger: " 
					     << daqdataformats::getTriggerShortName(trigType)
					     << std::endl;
	  else
	    ::mf::LogWarning("Trigger Rate") << "Specific Trigger Rate Average: " << info->Queue->getRate() 
					     << "\nSpecific Trigger rate exceeds maximum, throttling trigger output.\n" 
					     << "Type of last trigger: " 
					     << daqdataformats::getTriggerShortName(trigType)
					     << std::endl;
	
	  // We back off and cowardly refuse to issue the trigger
	  // We pop the lastest trigger (the one we are not going to issue) off the throttle queue
	  // which has the effect of pushing the rolling average down the next time through 
	  DDTTrigThrottle.popFront();
	  info->Queue->popFront();

	  // We bounce out since we aren't going to issue this trigger

	  info->incMissedCount();

	  continue;
	}
		

	// report the ddt time
	char ctimestr[27];ctime_r(&trigTime,ctimestr);        
	if (!(info->getSeenCount()%logLevel))
	  LOG_DEBUG("Progress") << "Received DataDrivenTrigger at time "
				<< ctimestr
				<< " with NovaTS=0x"<<hex<<ddtNova<<dec;

	// create a trigger atom
	// Version 2 is after 8/12/11's fleshing out of TriggerDefines.h
	daqdataformats::RawTrigger trigger((daqdataformats::version_t) 2);

	// Next, we construct the different segments of the trigger atom
	// one at a time and stuff them into the atom

	// Start with header
	daqdataformats::RawTriggerHeader triggerHeader;
      
	// Set HW trigger flag
	triggerHeader.setHWFlag(0); // calling DataDrivenDtriggers software

	// Set source
	triggerHeader.setSource(daqdataformats::TRIGSRC_BUFFNODE_SYSTEM);
	// Set Trigger Source based on where the ddtMessage is from
	triggerHeader.setSourceID(ddtMessage.buffer_node_id);
	triggerHeader.setSourceSubID(ddtMessage.process_id);

	// pack our header into the trigger atom later
	trigger.setTriggerHeader(triggerHeader);
	// Set trigger time.  
	daqdataformats::RawTriggerTime triggerTime;

	// This is the time the trigger was generated
	timeval now;
	gettimeofday(&now,0);
	triggerTime.setTriggerTimeStampTv(now);
	uint64_t ts_now;
	novadaq::timeutils::convertUnixTimeToNovaTime(now,ts_now);
	triggerTime.setTriggerTimeStamp(ts_now);
	triggerTime.setClockSource(daqdataformats::CLOCKSRC_GPS_FNAL_MINOS_2);

	// log DDT ganglia stats
	// cast length to 32 bits since ganglia doesn't like 64
	info->logDDT((uint32_t)ddtMessage.duration,
		     ddtMessage.time_to_decision,
		     (uint32_t)(ts_now - ddtMessage.start_time));

	// pack our time into the trigger atom
	trigger.setTriggerTime(triggerTime);

	// Set trigger time marker
	// This is the actual time used to go searching for data
	daqdataformats::RawTriggerTimingMarker triggerTimingMarker;
	uint64_t trigTimeNova=ddtMessage.start_time;

	triggerTimingMarker.setTriggerStart_baseClock(trigTimeNova);    
	// pack our time marker into the trigger atom
	trigger.setTriggerTimingMarker(triggerTimingMarker);

	// Set trigger mask
	daqdataformats::RawTriggerMask triggerMask;
	triggerMask.setMonteCarloFlag(0);  
	triggerMask.setTrigMod(0);
	// confusingly named method, same as the parent's method that 
	// sets this whole thing.  But, this takes a trigger bit number,
        // (trigType-1 here) and sets all the appropriate higher bits if needed
	triggerMask.setTriggerMask(info->getType()-1);
	
	// set prescale factor.  This is the actual prescale, so needs
	// to be multiplied by whatever the buffer node has already
	// applied
	if (ddtMessage.prescale > 0)   // try to trap unfilled bogosity
	  triggerMask.setPrescale(info->getPrescale()*ddtMessage.prescale);
	else
	  triggerMask.setPrescale(info->getPrescale());

	// Add in those we've missed since last trigger due to throttling.
	// The algebra is to account for the fact that if we throttled 
	// something that was already prescaled, we didn't just lose one 
	// more, but the last n as well
	triggerMask.setPrescale(triggerMask.getPrescale() * 
				(1 + info->getMissedCount()));
	info->resetMissedCount();
	// pack our mask into the trigger atom
	trigger.setTriggerMask(triggerMask);

	// let LT calculate info needed to chain the triggers        
	GTLongTrigs LT(ddtMessage.duration);
	// let LT adjust trigger times, loop over it and send while it
	// works
	bool first = true;
	while (LT.adjustTrigger(&trigger) > 0) {
	  // Set trigger number.
	  uint64_t TriggerNumberCopy = GT_Stats::getInstance().incTrigNo();
	  // keep track that we're sending a trigger of this type
          info->incSentCount();

	  triggerHeader.setTriggerNumber(TriggerNumberCopy);
	  if (first) {
	    first = false;
	    // Master Trigger number starts off the same
	    triggerHeader.setMasterTriggerNumber(TriggerNumberCopy);
	  }
	  // pack our header into the trigger atom - rest was loaded before
	  trigger.setTriggerHeader(triggerHeader);

	  // create the message, pack in the trigger atom, and send it
	  triggermessages::TriggerMessage theMessage;
	  theMessage.payload.length(4 * trigger.sizeofdata());
	  trigger.writeData(theMessage.payload.get_buffer());
	  // People are annoyed by this.  Rather than simply turning off
	  // debug messages, we've made logLevel into a message prescale.

	  if (!(info->getSeenCount()%logLevel))
	    LOG_DEBUG("Progress") << "Sending " << info->getName()
				  << " DDT trigger number " 
				  << TriggerNumberCopy <<endl;

	  // Queue the Message to the central sender queues
	  theQueues.QueueTrigger("DDT",theMessage);

        } // endwhile adjustTrigger
      } // endif prescale
    } // endwhile true
}

// Class Factories

// Class Create
extern "C" TriggerGenerator* createTrigger() {
  return new GTDataTrig;
}

// Class Destroy
extern "C" void destroyTrigger(TriggerGenerator* class_ptr){
  delete class_ptr;
}

