/* A thread dealing with Simulation Triggers */
/* The simulation package forms triggers completely, all we're doing
   here is forwarding them along */

#include <string>

#include <rms/provider/DDSConnection.h>
#include <rms/base/RmsDestination.h>
#include <rms/RmsSender.h>
#include <rms/RmsReceiver.h>

#include <messagefacility/MessageLogger/MessageLogger.h>

#include <DAQDataFormats/RawTriggerHeader.h>
#include <DAQDataFormats/RawTriggerTime.h>
#include <DAQDataFormats/RawTriggerTimingMarker.h>
#include <DAQDataFormats/RawTriggerRange.h>
#include <DAQDataFormats/RawTriggerMask.h>
#include <DAQDataFormats/RawTrigger.h>

#include <DAQDataFormats/TriggerDefines.h>

#include <NovaTimingUtilities/TimingUtilities.h>

#include <NovaGlobalTrigger/GT_SendQueues.h>

#include <DAQMessages/ccpp_GTMessages.h>
#include <NovaGlobalTrigger/GTGlobals.h>
#include <NovaGlobalTrigger/GT_Stats.h>

#include <boost/shared_ptr.hpp>

using namespace std;
using namespace gov::fnal::cd::rms;
using namespace novadaq;
using namespace gt;

#define TRIGGER_APP_NAME "GTSimTrig"

void GTSimTrig() {

    mf::SetModuleName("GTSimTrig");

    // Get the instance of the sending queues
    GT_SendQueues& theQueues = gt::GT_SendQueues::getInstance();

    // Get config stuff from global memory
    {
      boost::mutex::scoped_lock lock(gtCfgSet_mutex);

      // Grab the trigger config from global memory
      // This has to be there, at least one trigger is mandatory
      const TriggerType trigCfg = gtCfgSet.getGTConfiguration().getTriggers();

      // If there, go get the config info
      // Enable and prescale are the only things we keep track of
      // copy stuff to the GTInfo element for this trigger
      if (trigCfg.hasSimSet()) {
        SimType gtSimCfg = trigCfg.getSimSet().getSim();
        // First, bail if not enabled
        if (!gtSimCfg.getEnableFlag()) return;

        // We're still here?  Configure trigger parameters
	GTInfo *info = &GT_Stats::getInstance().Triggers.at(daqdataformats::TRIG_ID_SOFT_MONTECARLO);
	info->setEnabled(true);
        info->setPrescale(gtSimCfg.getPrescale());
	info->setName(gtSimCfg.getName());
	info->setType(daqdataformats::TRIG_ID_SOFT_MONTECARLO);
	info->initMetric();
      } else { 
	return; // bail if not configured
      }
    }  // end of mutexed config access

    ::mf::LogInfo("Progress") << "Simulation Trigger named " 
      << GT_Stats::getInstance().Triggers.at(daqdataformats::TRIG_ID_SOFT_MONTECARLO).getName() 
      << " enabled" << endl;

    // create the connection to RMS for the trigger
    boost::shared_ptr<provider::DDSConnection>
        rmsConnection(new provider::DDSConnection(TRIGGER_APP_NAME,partNumber));
    // create the destination to represent where the messages will be sent
    base::RmsDestination sendDest(TRIGGER_MAILBOX,
                                  base::RmsDestination::TRIGGER_CHANNEL);
    // create a sender with the connection and destination
    RmsSender<provider::DDSConnection,
        triggermessages::TriggerMessage>
        requestSender(rmsConnection, sendDest);

    // create the destination that represents where we will receive
    // messages from
    base::RmsDestination receiveDest("SimTriggerMailbox",
                                     base::RmsDestination::TRIGGER_CHANNEL);

    // create a receiver with the connection object and destination
    RmsReceiver<provider::DDSConnection,
        triggermessages::TriggerMessage>
        requestReceiver(rmsConnection, receiveDest);


    // loop over Simulated triggers
    uint64_t simTriggerCount = 0;
    while (1) {
      // If we're not triggering, wait
      while (!Triggering) {
        boost::mutex::scoped_lock waitlock(Wait_mutex);
        Triggering_cond.wait(waitlock);
      }

      triggermessages::TriggerMessage triggerMessage;
      requestReceiver.receiveMessage(triggerMessage);
      // copy the message data to a trigger atom
      daqdataformats::RawTrigger trigger;
      trigger.readData(triggerMessage.payload.get_buffer());

      GTInfo *info = &GT_Stats::getInstance().Triggers.at(daqdataformats::TRIG_ID_SOFT_MONTECARLO);

      // Apply prescale
      if (!(info->incSeenCount() % info->getPrescale()))
      {

      // Get the times out
      daqdataformats::RawTriggerTimingMarker* triggerTimingMarker = trigger.getTriggerTimingMarker();
      daqdataformats::RawTriggerTime* triggerTime = trigger.getTriggerTime();

      // We want to set the times in the TriggerTime with now, since
      // that's when the trigger was "generated".  We pass along the
      // TriggerTimingMarker, since that's when the sim wants to pull
      // actual data
      timeval now;
      gettimeofday(&now,0);
      uint64_t ts_now;
      novadaq::timeutils::convertUnixTimeToNovaTime(now,ts_now);
      triggerTime->setTriggerTimeStampTv(now);
      triggerTime->setTriggerTimeStamp(ts_now);

      timeval trigTime;
      uint64_t trigTimeNova;
      trigTimeNova = triggerTimingMarker->getTriggerStart_baseClock();
      novadaq::timeutils::convertNovaTimeToUnixTime(trigTimeNova,trigTime);

      // report the trigger time
      char ctimestr_now[27];ctime_r(&now.tv_sec,ctimestr_now);
      char ctimestr[27];ctime_r(&trigTime.tv_sec,ctimestr);
      if (!(simTriggerCount%logLevel))
        LOG_DEBUG("Progress") << "Received simulated trigger at time "
         << ctimestr_now 
         << " and " << now.tv_usec << " usec\n"
	 << "for a trigger occuring at simulated time " 
	 << ctimestr
	 << " and " << trigTime.tv_usec << " usec" << endl;

      // Set trigger number.  Under mutex protection.
      // Should reserve it first.

      daqdataformats::RawTriggerHeader* triggerHeader = trigger.getTriggerHeader();
      
      // Set trigger number.
      uint64_t TriggerNumberCopy = GT_Stats::getInstance().incTrigNo();
      // keep track that we're sending a trigger of this type
      info->incSentCount();
      triggerHeader->setTriggerNumber(TriggerNumberCopy);
      // Master Trigger number starts off the same, for short triggers
      triggerHeader->setMasterTriggerNumber(TriggerNumberCopy);
    
      // we don't need to pack our header into the trigger atom here,
      // since we're already twiddling stuff in there by using pointers
      // in rather than making new objects.  
      // For a simulated trigger, we change only the trigger number
      // generation time

      // create the message, pack in the trigger atom, and send it
      triggermessages::TriggerMessage theMessage;
      theMessage.payload.length(4 * trigger.sizeofdata());
      trigger.writeData(theMessage.payload.get_buffer());
      if (!(info->getSeenCount()%logLevel))
        LOG_DEBUG("Progress") << "Sending Spill trigger number " 
	                      << TriggerNumberCopy << endl;
      requestSender.sendMessage(theMessage);

      // Queue the Message to the central sender queues
      theQueues.QueueTrigger("Other",theMessage);

    } // endif prescale
    } 

    // cleanup and exit
    requestSender.close();
    rmsConnection->close();
}
