/* A thread dealing with Spill Triggers */
/* Triggers come from the same SpillServer message socket, we need to
   sort out which source (NuMI, BNB, or PPS) it is and deal with that. */

#include <string>

#include <rms/provider/DDSConnection.h>
#include <rms/base/RmsDestination.h>
#include <rms/RmsReceiver.h>

#include <messagefacility/MessageLogger/MessageLogger.h>

#include <DAQDataFormats/RawTriggerHeader.h>
#include <DAQDataFormats/RawTriggerTime.h>
#include <DAQDataFormats/RawTriggerTimingMarker.h>
#include <DAQDataFormats/RawTriggerRange.h>
#include <DAQDataFormats/RawTriggerMask.h>
#include <DAQDataFormats/RawTrigger.h>
#include <DAQDataFormats/TriggerDefines.h>

// Class Def
#include <NovaGlobalTrigger/GTSpillTrig.h>


#include <NovaTimingUtilities/TimingUtilities.h>

#include <NovaSpillServer/NssSpillInfo.h>
#include <DAQMessages/ccpp_NssMessages.h>

#include <DAQMessages/ccpp_GTMessages.h>
#include <NovaGlobalTrigger/GTGlobals.h>
#include <NovaGlobalTrigger/GT_Stats.h>
#include <NovaGlobalTrigger/GTLongTrigs.h>

#include <NovaGlobalTrigger/GT_SendQueues.h>

#include <boost/shared_ptr.hpp>

using namespace std;
using namespace gov::fnal::cd::rms;
using namespace novadaq;
using namespace gt;
using namespace daqdataformats;

#define TRIGGER_APP_NAME "GTSpillTrig"

void GTSpillTrig::configure(){
  setname(std::string("spill trigger"));
  return;
}

void GTSpillTrig::work() {

    mf::SetModuleName("GTSpillTrig");

    // Get the instance of the sending queues
    GT_SendQueues& theQueues = gt::GT_SendQueues::getInstance();


    {
      boost::mutex::scoped_lock lock(gtCfgSet_mutex);

      // Grab the trigger config from global memory
      // This has to be there, at least one trigger is mandatory
      const TriggerType trigCfg = gtCfgSet.getGTConfiguration().getTriggers();

      // Get Numi config stuff from global memory if it is there
      // Note: need to trap on the potential lack of stuff inside the Set
      // an empty Set could break this
      if (trigCfg.hasNumiSpillSet()) {
        NumiSpillType gtNumiSpillCfg = trigCfg.getNumiSpillSet().getNumiSpill();
	// copy stuff to the GTInfo element for this trigger: bit #0
	GTInfo *info = &GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_NUMI);
	info->setEnabled(gtNumiSpillCfg.getEnableFlag());
	info->setLength(gtNumiSpillCfg.getLength());
	info->setPrescale(gtNumiSpillCfg.getPrescale());
	info->setDelay(gtNumiSpillCfg.getDelay());
	info->setName(gtNumiSpillCfg.getName());
	info->setType(TRIG_ID_BEAM_NUMI);
	if (info->getEnabled()) info->initMetric();
      }
      if (trigCfg.hasMinibooneSpillSet()) {
        // Get BNB config stuff from global memory if it is there
        MinibooneSpillType gtMinibooneSpillCfg = trigCfg.getMinibooneSpillSet().getMinibooneSpill();
	// copy stuff to the GTInfo element for this trigger: bit #1
	GTInfo *info = &GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_BOONE);
	info->setEnabled(gtMinibooneSpillCfg.getEnableFlag());
	info->setLength(gtMinibooneSpillCfg.getLength());
	info->setPrescale(gtMinibooneSpillCfg.getPrescale());
	info->setDelay(gtMinibooneSpillCfg.getDelay());
	info->setName(gtMinibooneSpillCfg.getName());
	info->setType(TRIG_ID_BEAM_BOONE);
	if (info->getEnabled()) info->initMetric();
      }
      if (trigCfg.hasPpsSpillSet()) {
        // Get PPS config stuff from global memory if it is there
        PpsSpillType gtPpsSpillCfg = trigCfg.getPpsSpillSet().getPpsSpill();
	// copy stuff to the GTInfo element for this trigger: bit #5
	GTInfo *info = &GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_1PPS);
	info->setEnabled(gtPpsSpillCfg.getEnableFlag());
	info->setLength(gtPpsSpillCfg.getLength());
	info->setPrescale(gtPpsSpillCfg.getPrescale());
	info->setDelay(gtPpsSpillCfg.getDelay());
	info->setName(gtPpsSpillCfg.getName());
	info->setType(TRIG_ID_BEAM_1PPS);
	if (info->getEnabled()) info->initMetric();
      }
      // Bail if no spills enabled
      if (!(GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_NUMI).getEnabled()
            ||GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_BOONE).getEnabled()
	    ||GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_1PPS).getEnabled())) return;

      // create overall spill metrics only if we reach here, thus at least 
      // one spill trigger must be enabled
      GT_Stats::getInstance().initSpillMetrics();
    }  // end of mutexed config access

    // Log which spills are enabled
    if (GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_NUMI).getEnabled()) 
      ::mf::LogInfo("Progress") << "NuMI Spill Trigger named "
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_NUMI).getName() 
	<< " enabled with length " 
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_NUMI).getLength()
        << ", delay " 
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_NUMI).getDelay()
        << ", and prescale " 
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_NUMI).getPrescale() 
	<< endl;

    if (GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_BOONE).getEnabled()) 
      ::mf::LogInfo("Progress") << "BNB Spill Trigger named "
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_BOONE).getName() 
	<< " enabled with length " 
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_BOONE).getLength()
        << ", delay " 
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_BOONE).getDelay()
        << ", and prescale " 
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_BOONE).getPrescale() 
	<< endl;

    if (GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_1PPS).getEnabled()) 
      ::mf::LogInfo("Progress") << "PPS Spill Trigger named "
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_1PPS).getName() 
	<< " enabled with length " 
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_1PPS).getLength()
        << ", delay " 
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_1PPS).getDelay()
        << ", and prescale " 
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_BEAM_1PPS).getPrescale() 
	<< endl;

    // create the RMS connection for the spills, on the null partition
    boost::shared_ptr<provider::DDSConnection>
        rmsSpillConnection(new provider::DDSConnection("GTSpillListener",
                                    base::RmsDestination::NULL_PARTITION));
    // create the destination that represents where we will receive
    // messages from
    base::RmsDestination receiveDest("SpillServerMailbox",
                                     base::RmsDestination::TRIGGER_CHANNEL);
    // create a receiver with the connection object and destination
    RmsReceiver<provider::DDSConnection,
        nssmessages::SpillMessage>
        requestReceiver(rmsSpillConnection, receiveDest);

    // loop over Spill triggers
    while (1) {
      // If we're not triggering, wait
      while (!Triggering) {
        boost::mutex::scoped_lock waitlock(Wait_mutex);
        Triggering_cond.wait(waitlock);
      }

      nova::NovaSpillServer::NssSpillInfo spillinfo;

      nssmessages::SpillMessage spillMessage;

      requestReceiver.receiveMessage(spillMessage);
      // copy the message data to a spillinfo object
      spillinfo.set(spillMessage.time,spillMessage.type);

      // Get which type of trigger it is
      trigID spillType = TRIG_ID_NULL;
      trigSourceID whichSignal = TRIGSRC_ID_BEAMSTAMP_UNKNOWN;
      if (spillMessage.type==nova::NovaSpillServer::kNuMI||
          spillMessage.type==nova::NovaSpillServer::kNuMItclk) {
	  spillType = TRIG_ID_BEAM_NUMI;
	  if (spillMessage.type==nova::NovaSpillServer::kNuMI)
	     // a kNuMI trigger
	     whichSignal = TRIGSRC_ID_BEAMSTAMP_NUMI_1;
	  else
	     // a kNuMItclk trigger
	     whichSignal = TRIGSRC_ID_BEAMSTAMP_NUMI_2;
      }
      if (spillMessage.type==nova::NovaSpillServer::kBNB||
          spillMessage.type==nova::NovaSpillServer::kBNBtclk) {
	  spillType = TRIG_ID_BEAM_BOONE;
	  if (spillMessage.type==nova::NovaSpillServer::kBNB)
	     // a kBNB trigger
	     whichSignal = TRIGSRC_ID_BEAMSTAMP_BOOSTER_1;
	  else
	     // a kBNBtclk trigger
	     whichSignal = TRIGSRC_ID_BEAMSTAMP_BOOSTER_2;
      }
      if (spillMessage.type==nova::NovaSpillServer::kAccelOneHztclk) {
	  spillType = TRIG_ID_BEAM_1PPS;
	  whichSignal = TRIGSRC_ID_BEAMSTAMP_PPS;
      }
      if (spillType==TRIG_ID_NULL) {
	 ::mf::LogWarning("Unknown Spill") << "Spillserver has sent an unknown spill type, ID " << spillMessage.type;
      }
      GTInfo *info = &GT_Stats::getInstance().Triggers.at(spillType);

      // If that trigger isn't enabled skip to wait for next spill
      if (!info->getEnabled()) continue;

      // Apply prescale
      if (!(info->incSeenCount() % info->getPrescale())) {
        // Get the time out, convert back to t_time
        uint64_t spillNova = spillinfo.getTime();
        timeval spillTv;
        novadaq::timeutils::convertNovaTimeToUnixTime(spillNova,spillTv);
        time_t trigTime = spillTv.tv_sec;

        // report the spill time
        char ctimestr[27];ctime_r(&trigTime,ctimestr);
	if (!(info->getSeenCount()%logLevel))
	   LOG_DEBUG("Progress") << "Received " << info->getName()
	     << " spill at time "
             << ctimestr << " with NovaTS=0x"<<hex<<spillNova<<dec;

        // create a trigger atom
	// Version 2 is after 8/12/11's fleshing out of TriggerDefines.h
        RawTrigger trigger((version_t) 2);

        // Next, we construct the different segments of the trigger atom
        // one at a time and stuff them into the atom

        // Start with header
        RawTriggerHeader triggerHeader;
      
        // Set HW trigger flag
        triggerHeader.setHWFlag(1); // calling beam a HW trigger

        // Set source
        triggerHeader.setSource(TRIGSRC_BEAMSTAMP_SYSTEM);
        triggerHeader.setSourceID(whichSignal);
	// atm, the TDU stamping all spills is in the MINOS
	// surface building, the secondary TDU
        triggerHeader.setSourceSubID(TRIGSRC_SUBID_TDU_02);
 
        // Set trigger time.  
        RawTriggerTime triggerTime;

        // This is the time the trigger was generated
        timeval now;
        gettimeofday(&now,0);
        triggerTime.setTriggerTimeStampTv(now);
        uint64_t ts_now;
        novadaq::timeutils::convertUnixTimeToNovaTime(now,ts_now);
        triggerTime.setTriggerTimeStamp(ts_now);

        triggerTime.setClockSource(CLOCKSRC_GPS_FNAL_MINOS_2);

        // Issue a warning if the spill is "too old".  Manually defined as
        // off from current time by more than a minute, or in the future.
        // Will change this to something configurable later.
        uint64_t deltat = ts_now - spillNova;
	double deltat_s = (double)deltat/(double)timeutils::NOVA_TIME_FACTOR;
	GT_Stats::getInstance().logSpillDelay(deltat_s);
        if (spillNova < ts_now) {
	  if (deltat_s > 60)
	    ::mf::LogWarning("Time Shear") << "Spill time is much older than current time " << deltat_s <<" s";
        } else {
          ::mf::LogWarning("Time Shear") << "Spill time is in the future " << deltat_s*(-1000.0) << " ms";
        }

        // pack our time into the trigger atom
        trigger.setTriggerTime(triggerTime);

        // Set trigger time marker
        // This is the actual spill time used to go searching for data
        RawTriggerTimingMarker triggerTimingMarker;
        uint64_t trigTimeNova=spillNova;
        trigTimeNova+=info->getDelay();

        triggerTimingMarker.setTriggerStart_baseClock(trigTimeNova);    
        // pack our time marker into the trigger atom
        trigger.setTriggerTimingMarker(triggerTimingMarker);

        // Set trigger mask
        RawTriggerMask triggerMask;
        triggerMask.setMonteCarloFlag(0);  // Not MC
        triggerMask.setTrigMod(0);         // This is the "Trigger Set"
        // confusingly named method, same as the parent's method that
        // sets this whole thing.  But, this takes a trigger bit number,
        // (trigType-1 here) and sets all the appropriate higher bits if
        // needed
        triggerMask.setTriggerMask(info->getType()-1);
	// set prescale factor
        triggerMask.setPrescale(info->getPrescale());

        // pack our mask into the trigger atom
        trigger.setTriggerMask(triggerMask);

        // let LT calculate info needed to chain the triggers        
        GTLongTrigs LT(info->getLength());
        // let LT adjust trigger times, loop over it and send while it
        // works
        bool first = true;
        while (LT.adjustTrigger(&trigger) > 0) {

	  // Set trigger number.
	  uint64_t TriggerNumberCopy = GT_Stats::getInstance().incTrigNo();
	  // keep track that we're sending a trigger of this type
	  info->incSentCount();
	  triggerHeader.setTriggerNumber(TriggerNumberCopy);
	  if (first) {
	    first = false;
            // Master Trigger number starts off the same
            triggerHeader.setMasterTriggerNumber(TriggerNumberCopy);
	  }
	  // pack our header into the trigger atom - rest was loaded before
	  trigger.setTriggerHeader(triggerHeader);

	  // create the message, pack in the trigger atom, and send it
	  triggermessages::TriggerMessage theMessage;
	  theMessage.payload.length(4 * trigger.sizeofdata());
	  trigger.writeData(theMessage.payload.get_buffer());
	  // People are annoyed by this.  Rather than simply turning off
	  // debug messages, we've made logLevel into a message prescale.

	  if (!(info->getSeenCount()%logLevel))
            LOG_DEBUG("Progress") << "Sending " << info->getName() 
	      << " Spill trigger number " << TriggerNumberCopy 
//              <<" with NovaTS=0x"<<hex
//              <<triggerTimingMarker.getTriggerStart_baseClock()
//              <<" with GenTime(Nova)=0x"
//              <<trigger.getTriggerTime()->getTriggerTimeStamp()
//              << dec
//              <<" (gen timeval = "<<now.tv_sec
//              <<" sec, "<<now.tv_usec
//              <<" usec.)"
              <<endl;
	  
	  // Queue the Message to the central sender queues
	  theQueues.QueueTrigger("Beam",theMessage);

        }
      } // endif prescale
    } 
}
