/* A thread listening for SuperNova dds messages */

#include <string>

#include <rms/provider/DDSConnection.h>
#include <rms/base/RmsDestination.h>
#include <rms/RmsReceiver.h>

#include <messagefacility/MessageLogger/MessageLogger.h>

#include <DAQDataFormats/TriggerDefines.h>

#include <NovaTimingUtilities/TimingUtilities.h>

#include <DAQMessages/ccpp_NSNMessages.h>

#include <NovaGlobalTrigger/GTGlobals.h>
#include <NovaGlobalTrigger/GT_Stats.h>
#include <NovaGlobalTrigger/GTLongTrigs.h>

#include <NovaGlobalTrigger/GTSuperNovaListener.h>
#include <NovaSuperNova/DDSMailbox.hpp>
//#include <NovaGlobalTrigger/GT_SendQueues.h>


using namespace std;
using namespace gov::fnal::cd::rms;
using namespace novadaq;
using namespace gt;
using namespace daqdataformats;

#define TRIGGER_APP_NAME "GTSuperNovaListener"

std::ostream& operator<<(std::ostream& stream, supernova::DataBuffer& buf){
   stream<<"DataBuffer: t=["<<buf.startTime()<<"-"<<buf.endTime()<<" duration="<<buf.duration()
         <<"With "<<buf.size()<<" blocks"<<std::endl;
   return stream;
}

bool IsEndOfOperationMsg(const nsnmessages::NSNPackedMessage& msg){
   //check if this message is completely empty
   for(size_t n=0; n<10;++n){
      if(msg.time[n]!=0)return false;
      if(msg.hits[n]!=0)return false;
   }
   return true;
}

void GTSuperNovaListener::configure() { 
    mf::SetModuleName("GTSuperNovaListener");
    setname(std::string("data driven trigger")); 

    { 
      boost::mutex::scoped_lock lock(gtCfgSet_mutex);
      // Grab the trigger config from global memory
      // This has to be there, at least one trigger is mandatory
      const TriggerType trigCfg = gtCfgSet.getGTConfiguration().getTriggers();
      configured = false;
      // If there, go get the config info
      if (!trigCfg.hasSuperNovaSet()) return;
      if (trigCfg.getSuperNovaSet().hasDataBuffer()){
          DataBufferType bufferCfg=trigCfg.getSuperNovaSet().getDataBuffer();
          size_t nbins=bufferCfg.getBinsPerBlock();
          supernova::TimeType dtbin=bufferCfg.getBinDuration();
          inBuffer.setBlocksLength  (nbins);
          inBuffer.setBlocksDuration(nbins*dtbin);
          _blockTimeout=bufferCfg.getBlockTimeout();
          ::mf::LogInfo("Config")
           <<"Set DataBlock[nbins="<<nbins<<" time="<<nbins*dtbin
           <<" (dtBin="<<dtbin<<")]"<<std::endl
           <<"blockTimeout="<<_blockTimeout<<std::endl;
      } 
      if (trigCfg.getSuperNovaSet().hasDDSInbox()){
          _ddsPartition=trigCfg.getSuperNovaSet().getDDSInbox().getPartition();
          _ddsBoxName  =trigCfg.getSuperNovaSet().getDDSInbox().getName();
        ::mf::LogInfo("Config")<<"Setup DDS ["<<_ddsBoxName<<", part#"<<_ddsPartition<<"]"<<std::endl;
      }
      if (!trigCfg.getSuperNovaSet().hasSNTrigger()) return;
      SNTriggerType gtSNTriggerCfg = trigCfg.getSuperNovaSet().getSNTrigger();
      
      // First, bail if not enabled
      if (!gtSNTriggerCfg.getEnableFlag()) return;
      //if we've reached this point, then configuration was successfull
      configured = true;
    }
}// end configure

void GTSuperNovaListener::work() {
    if(!configured){
      ::mf::LogWarning("Work")<<"SuperNova trigger was not configured!"<<std::endl;
      return;
    }
    nsnmessages::NSNPackedMessage message;
    DDSInbox<nsnmessages::NSNPackedMessage> ddsbox("SuperNovaMailbox",_ddsPartition,_ddsBoxName.data());
    ::mf::LogInfo("Work")<<"Listening to DDS ["<<_ddsBoxName<<", part#"<<_ddsPartition<<"]"<<std::endl;
    while(true) {
         ddsbox>>message; //receive the message 
         ::mf::LogInfo("Progress")<<"Received NSNMessage: time="<<message.time[0]<<"-"<<message.time[9]<<std::endl;
         try{
            if(IsEndOfOperationMsg(message)){
                //special case: received message indictes end of operation
                //add empty block in the end
                inBuffer.addBlockFront(); 
                //dump all blocks to processing
                while(!inBuffer.empty())forwardNextBlock();
                return;
            }

            //TODO: discard expiredmessages
            for(size_t n=0;n<10;++n){
                //dump all expired blocks to processing buffer
                forwardExpiredBlocks(message.time[n]);
                //fill the buffer
                inBuffer.fill(message.time[n],message.hits[n]);               
               }
            forwardFilledBlocks();
         }
         catch(supernova::errors::basic_error& err){
            ::mf::LogError("Progress")<<err.what();//<<std::endl;
         }
         ::mf::LogInfo("Progress")<<"Buffer ["<<inBuffer.startTime()<<" : "<<inBuffer.endTime()<<"]="
         <<inBuffer.duration()<<"len="<<inBuffer.length()<<"/"<<inBuffer.size()<<std::endl; 
       }
}

void GTSuperNovaListener::forwardExpiredBlocks(supernova::TimeType currentTime){
    while(!inBuffer.empty() && inBuffer.backBlock().endTime()+_blockTimeout<currentTime){
        ::mf::LogInfo("Progress")<<"Pop OLD block. dt="<<currentTime-inBuffer.backBlock().endTime()<<" > "<<_blockTimeout<<std::endl;
        forwardNextBlock();
        }
}

void GTSuperNovaListener::forwardFilledBlocks(){
    while(!inBuffer.empty() && inBuffer.backBlock().full()){ 
        forwardNextBlock();
        ::mf::LogInfo("Progress")<<"Pop full block"<<std::endl;
        }
}

void GTSuperNovaListener::forwardNextBlock(){
    //move block from inBuffer to sharedBuffer;
     if(outBuffer_ptr!=0){
          boost::mutex::scoped_lock lock(*outBuffer_ptr); //lock outBuffer 
          outBuffer_ptr->pushBlockFront(inBuffer.backBlock()); //and write a block
      }
      inBuffer.popBlockBack();
}

