/* A thread listening for SuperNova dds messages */

#include <string>
#include <sstream>
#include <vector>
#include <messagefacility/MessageLogger/MessageLogger.h>

#include <DAQDataFormats/TriggerDefines.h>

#include <NovaTimingUtilities/TimingUtilities.h>

#include <NovaGlobalTrigger/GTGlobals.h>
#include <NovaGlobalTrigger/GT_Stats.h>
#include <NovaGlobalTrigger/GTLongTrigs.h>

#include <NovaGlobalTrigger/GTSuperNovaTrig.h>

//#include <NovaGlobalTrigger/GT_SendQueues.h>

using namespace novadaq;
using namespace gt;
using namespace daqdataformats;

supernova::FIRFilter MakeFilterFromString(std::string str){
    // read filter kernel points from given string
    std::istringstream iss(str);
    std::vector<double> points;
    double val;
    while(iss){
          if(iss>>val) points.push_back(val);
    };
    
    //create filter with needed kernel size
    supernova::FIRFilter filter(points.size());
    //fill kernel with values
    for(size_t i=0;i<points.size();++i)
        filter[i]=points[i];
    //filter is ready!  
    return filter;
}

GTSuperNovaTrig::~GTSuperNovaTrig(){
   _logfile.close();
}
void GTSuperNovaTrig::configure() { 
    mf::SetModuleName("GTSuperNovaTrigger");
    setname(std::string("data driven triggers")); 
    
    { 
      boost::mutex::scoped_lock lock(gtCfgSet_mutex);
      // Grab the trigger config from global memory
      // This has to be there, at least one trigger is mandatory
      const TriggerType trigCfg = gtCfgSet.getGTConfiguration().getTriggers();

      // If there, go get the config info
      if (!trigCfg.hasSuperNovaSet()) return;
      if (!trigCfg.getSuperNovaSet().hasSNTrigger())return;
      if (!trigCfg.getSuperNovaSet().getSNTrigger().getEnableFlag()) return;

      //open logfile
      if (trigCfg.getSuperNovaSet().hasLogFile()){
        std::string filename=trigCfg.getSuperNovaSet().getLogFile().getName();
       _logfile.open(filename.data(), std::ofstream::out); 
      }      
      //configure trigger itself 
      SuperNovaSetType::SNTrigger_type gtSNTriggerCfg = trigCfg.getSuperNovaSet().getSNTrigger();
      _threshold=gtSNTriggerCfg.getThreshold();
      ::mf::LogInfo("Config")<<"set threshold="<<_threshold;
      if(isLogging())_logfile<<"#TRIGGER_THRESHOLD "<<_threshold<<std::endl;

      if (trigCfg.getSuperNovaSet().hasFilter()){
        SuperNovaSetType::Filter_type filterCfg=trigCfg.getSuperNovaSet().getFilter();
        std::string name=filterCfg.getName();
        
        ::mf::LogInfo("Config")<<"Setup Filter \""<<filterCfg.getName()<<"\""
        << " with K=["<<filterCfg.getKernel()<<"]"<<std::endl;
        if(isLogging())
        _logfile<<"#FILTER \""<<filterCfg.getName()<<"\" K=["<<filterCfg.getKernel()<<"]"<<std::endl;
        ///configure filter
        filter=MakeFilterFromString(filterCfg.getKernel());
        if(filterCfg.getNormalize()) filter.normalize();
      }else{
        ::mf::LogError("Config")<<"No filter present in XML config!"<<std::endl;
      }

    }


}// end configure

void GTSuperNovaTrig::work() {
   if(isLogging())_logfile.width(20);
   while(true){
      //wait until we have a DataBlock in our buffer
       if(isReadyToProcess())
          Process();
   }
}

bool GTSuperNovaTrig::SignalIsFound(){
   ///TODO
    return false;
}

void GTSuperNovaTrig::SendTriggerSignal(supernova::TimeType time){
    ///TODO actual trigger signal sending
    _logfile<<"#TRIGGER_START "<<time<<std::endl;
}


bool GTSuperNovaTrig::isReadyToProcess(){
   boost::mutex::scoped_lock lock(buffer);
   //we need to have at least 2 blocks, so we can process the first one
   return (buffer.size()>1);
}

void GTSuperNovaTrig::Process(){
   //lock this buffer, so SuperNovaListener process doesn't modify it, while we process
   boost::mutex::scoped_lock lock(buffer);
   //take the last block
   supernova::DataBlock &block=buffer.backBlock();
   //std::cout<<"processing  block "<<block<<std::endl;
   //temporary: output the data and filtered signal
   supernova::TimeType time;
   double data, res;

   if(isLogging())_logfile<<"#block_start "<<block.startTime()<<" "<<block.endTime()<<std::endl;

   for(supernova::BufferIterator it=buffer.beginBlock(0);it!=buffer.endBlock(0);++it){
      time=it.timestamp();
      data=it.value();
      res=filter.convolve(it);
      if(isLogging())_logfile<<time<<" "<<data<<" "<<res<<std::endl;
      ///check if filtered signal is above threshold
      if(res>_threshold)SendTriggerSignal(time);
   }
   if(isLogging())_logfile.flush();
   ///remove processed block from buffer
   buffer.popBlockBack();
}
