/**
 * NGTMCStateMachine implementation file
 */
#include <sstream>
#include <messagefacility/MessageLogger/MessageLogger.h>
#include <NovaGlobalTrigger/NGTMCStateMachine.h>

using namespace std;
using namespace novadaq::gt;

/**
 * Constructor
 */
NGTMCStateMachine::NGTMCStateMachine() {

  LOG_DEBUG("NGTMCStateMachine") << "NGTMCStateMachine::ctor @" << this << "." << std::endl;
  
  for (int32_t i=0; i<NUMPARTS; i++) {
    _currentState[i]=NO_PARTITION;
  }
}

/**
 * Destructor
 */
NGTMCStateMachine::~NGTMCStateMachine() {

  LOG_DEBUG("NGTMCStateMachine") << "NGTMCStateMachine::dtor @" << this << "." << std::endl;

}

/**
 * Static method to return NGTMCState_t as a string.
 */
std::string NGTMCStateMachine::NGTMCStateAsString(NGTMCState_t state) {

  switch ( state ) {

  case NO_PARTITION:
    return "NO_PARTITION";

  case PARTITION_ESTABLISHED:
    return "PARTITION_ESTABLISHED";

  case RUN_CONFIGURATION_LOADED:
    return "RUN_CONFIGURATION_LOADED";

  case RUN_CONFIGURED:
    return "RUN_CONFIGURED";

  default:
    ::mf::LogError("NGTMCStateMachine") << "NGTMCStateMachine::NGTMCStateAsString - received unknown state " 
                                << state << "." << std::endl;
  }

  return "UNKNOWN";
}


/** 
 * Method to return the current state of a partition
 */
NGTMCStateMachine::NGTMCState_t NGTMCStateMachine::getCurrentState(int32_t partition) { 
     return _currentState[partition]; 
  }

/**
 * Method to change from _currentState to newState.  If a transition isn't allowed,
 * or if the transition fails, the method returns false.  
 * Transitions must be requested incrementally when moving upward through the state transitions,
 * but larger transition jumps may be requested when moving downward through the
 * state transitions.  (The intermediate downward transition steps will be handled internally.)
 * The reason for this difference in behavior is that many of the upward state transitions
 * require configuration information, while downward transitions do not.
 *  
 * @param newState
 * @param configURI optional string to configuration file
 * @param partition number (def = -1)
 * @return true if success, else false.  
 */
bool NGTMCStateMachine::changeState(NGTMCState_t newState, std::string configURI, int32_t partition) {

  std::string loginfo = "NGTMCStateMachine::changeState - request to move to state "
    + NGTMCStateAsString(newState) + " from state " 
    + NGTMCStateAsString(_currentState[partition]);
  if ( !configURI.empty() ) loginfo += " with configURI " + configURI;
  if ( partition != -1 ) {
    std::stringstream partitionstr;
    partitionstr << partition;
    loginfo += ", partition " + partitionstr.str();
  }
  ::mf::LogInfo("NGTMCStateMachine") << loginfo.c_str() << std::endl;

  if ( _currentState[partition] == NO_PARTITION ) {
    return _changeStateFromNoPartition(newState,configURI,partition);
  }
  else if ( _currentState[partition] == PARTITION_ESTABLISHED ) {
    return _changeStateFromPartitionEstablished(newState,configURI,partition);
  }
  else if ( _currentState[partition] == RUN_CONFIGURATION_LOADED ) {
    return _changeStateFromRunConfigurationLoaded(newState,configURI,partition);
  }
  else if ( _currentState[partition] == RUN_CONFIGURED ) {
    return _changeStateFromRunConfigured(newState,configURI,partition);
  }
  
  ::mf::LogError("NGTMCStateMachine") << "NGTMCStateMachine::changeState - Unknown current state "
                              << _currentState[partition] << "??? Fail." << std::endl;
  return false;

}

/**
 * Private method to change state when current state is NO_PARTITION.
 * @param newState
 * @param configURI optional string to configuration file
 * @param partition number
 * @return true if success, else false.  
 */
bool NGTMCStateMachine::_changeStateFromNoPartition(NGTMCState_t newState, 
                    std::string configURI, int32_t partition) {

  switch ( newState ) {

  case NO_PARTITION:
    return true;

  case PARTITION_ESTABLISHED:
    if (_establishPartition(partition) ) {
      _currentState[partition] = newState;
      return true;
    }
    else return false;

  default:
    ::mf::LogError("NGTMCStateMachine") << "NGTMCStateMachine::_changeState - request to jump to state " 
                                << NGTMCStateAsString(newState) << " from state " 
                                << NGTMCStateAsString(_currentState[partition]) << " denied." << std::endl;
    return false;
  }

}


/**
 * Private method to change state when current state is PARTITION_ESTABLISHED.
 * @param newState
 * @param configURI optional string to configuration file
 * @return true if success, else false.  
 */
bool NGTMCStateMachine::_changeStateFromPartitionEstablished(NGTMCState_t newState, 
                                                           std::string configURI, 
                                                           int32_t partition) {

  switch ( newState ) {

  case NO_PARTITION:
    if (_detachFromPartition(partition)) {
      _currentState[partition] = NO_PARTITION;
      return true;
    } else return false;

  case PARTITION_ESTABLISHED:
    if ( partition != -1 ) {
      ::mf::LogError("NGTMCStateMachine") 
                  << "NGTMCStateMachine::_changeStateFromPartitionEstablished - Unable to establish"
		  << " new partition " << partition 
                  << " without release from established partition first.  "
                  << "Request denied." << std::endl;
      return false;
    }
    else return true;

  case RUN_CONFIGURATION_LOADED:
    if ( _loadRunConfiguration(configURI,partition) ) {
      _currentState[partition] = newState;
      return true;
    }
    else return false;

  default:
    ::mf::LogError("NGTMCStateMachine") << "NGTMCStateMachine::_changeState - request to jump to state " 
                                << NGTMCStateAsString(newState) << " from state " 
                                << NGTMCStateAsString(_currentState[partition]) << " denied." << std::endl;
    return false;
  }

}


/**
 * Private method to change state when current state is RUN_CONFIGURATION_LOADED.
 * @param newState
 * @param configURI optional string to configuration file
 * @return true if success, else false.  
 */
bool NGTMCStateMachine::_changeStateFromRunConfigurationLoaded(NGTMCState_t newState, 
                        std::string configURI, int32_t partition) {

  switch ( newState ) {

  case NO_PARTITION:
  case PARTITION_ESTABLISHED:
    _unloadRunConfiguration(partition);
    _currentState[partition] = PARTITION_ESTABLISHED;
    return changeState(newState,configURI,partition);

  case RUN_CONFIGURATION_LOADED:
    if ( !configURI.empty() ) {
      _unloadRunConfiguration(partition);
      _currentState[partition] = PARTITION_ESTABLISHED;
      return changeState(newState,configURI,partition);
    }
    else return true;

  case RUN_CONFIGURED:
    if ( _configureRun(partition) ) {
      _currentState[partition] = newState;
      return true;
    }
    else return false;

  default:
    ::mf::LogError("NGTMCStateMachine") << "NGTMCStateMachine::_changeState - request to jump to state " 
                                << NGTMCStateAsString(newState) << " from state " 
                                << NGTMCStateAsString(_currentState[partition]) << " denied." << std::endl;
    return false;
  }

}

/**
 * Private method to change state when current state is RUN_CONFIGURED.
 * @param newState
 * @param configURI optional string to configuration file
 * @return true if success, else false.  
 */
bool NGTMCStateMachine::_changeStateFromRunConfigured(NGTMCState_t newState, std::string configURI, int32_t partition) {

  switch ( newState ) {

  case NO_PARTITION:
  case PARTITION_ESTABLISHED:
  case RUN_CONFIGURATION_LOADED:
    if (_deconfigureRun(partition)) {
      _currentState[partition] = RUN_CONFIGURATION_LOADED;
      return changeState(newState, configURI, partition);
    } else return false;

  case RUN_CONFIGURED:
    return true;

  default:
    ::mf::LogError("NGTMCStateMachine") << "NGTMCStateMachine::_changeState - request to jump to state " 
                                << NGTMCStateAsString(newState) << " from state " 
                                << NGTMCStateAsString(_currentState[partition]) << " denied." << std::endl;
    return false;
  }

}
