#include <sys/types.h>
#include <sys/wait.h>

#include <boost/thread/thread.hpp>

#include <NovaGlobalTrigger/NGTMasterRC.h>

#include <rms/base/RmsDestination.h>
#include <messagefacility/MessageLogger/MessageLogger.h>

// Should not need these!!
//#define DEFINE_GLOBALS
//#include <NovaGlobalTrigger/GTGlobals.h>

#include <NovaGlobalTrigger/GT_Inst.h>

// for cmd-line parsing
#include <boost/program_options.hpp>
namespace po = boost::program_options;

using namespace std;
using namespace gov::fnal::cd::rms;
using namespace novadaq;

static void sigchld_hdl(int32_t);  // SIGCHLD handerler prototype

int gPartition;
int verbose;

//------------------------------------------------------------

void PrintUsage()
{
  std::cout << "NGTMaster -z (--partition) [number >= 0]" << std::endl;
}

//------------------------------------------------------------

bool ParseCLArgs(int ac, char* av[])
{

  // Parse command line options using Boost
  // Declare the supported options
  po::options_description desc("Allowed options");
  desc.add_options()
    ("help,h", "usage descriptions")
    ("zpartition,z", po::value<int>(&gPartition)->default_value(0), "the partition to talk to, new form")
    ("verbose,v", po::value<int>(&verbose)->default_value(0), "verbosity level")
    ;
  po::variables_map vm;
  po::store(po::parse_command_line(ac, av, desc), vm);
  po::notify(vm);
  if (vm.count("help")) {
    std::cout << desc << "\n";
    exit(1);
  }
  
  return true;
}


//------------------------------------------------------------

int main(int argc, char* argv[]) {

  ParseCLArgs(argc,argv);

  // Attach to message logging service here. 
  // this uses the multi-threaded version, logs to console, and to 
  // the message facility on the NULL partition
  std::string msgFacCfgFile = "GlobalTriggerMsgFac.fcl";
  mf::StartMessageFacility( mf::MessageFacilityService::MultiThread, 
			    mf::MessageFacilityService::
                            ConfigurationFile(msgFacCfgFile.c_str(),
                                              mf::MessageFacilityService::
                                              logCS()));
  mf::SetApplicationName("NGTMaster");
  mf::SetModuleName("main");
  //  mf::SetContext("pre-event");  // what does this do?

  // set up signal handler
  struct sigaction act;
  memset (&act, 0, sizeof(act));
  act.sa_flags = SA_RESTART | SA_NOCLDSTOP;
  act.sa_handler = sigchld_hdl;
  if (sigaction(SIGCHLD, &act, NULL)) {
    ::mf::LogError("Progress") << "Can't set up SIGCHLD handler" << endl;
    return 1;
  }

  // create the GlobalTrigger application and start it
  // here, GlobalTrigger is the name rc uses to talk to us
  gt::NGTMasterRC NGTMasterRC("GlobalTrigger",gPartition);
  boost::thread gtMCRCThread(boost::bind(&novadaq::Runnable::run, &NGTMasterRC));

  // Scheme: make an array of pid's, one per partition.  Init to 0's
  // have create partition handler check if one's made already, if not, 
  // make one by fork/exec ing a NovaGlobalTrigger.
  // CAVEAT - need to have a configure interaction first, store xml
  // to a file, so NovaGlobalTrigger can read it.
  // For now, just force partition number on the command line

  // For release partitions, send a signal to the pid, check that it's dead
  // then return.

  // All run start, stop, pause, resume interactions are done by 
  // NovaGlobalTrigger.
  // That said, we should have the handlers here, so that we can 
  // complain if told to do it out of state.

  // Create the singleton instance.  Doing it here since doing it later
  // in a thread seems to make multiple copies.
  gt::GT_Inst::getInstance();

  // Wait for RC thread to end.  Do we ever really want to quit this 
  // program short of a KILL?
  gtMCRCThread.join();

  return 0;
}


/* SIGCHLD handler. */
static void sigchld_hdl(int32_t sig) {

  pid_t pid;
  int32_t status;
  int32_t partno;

  /* Wait for a dead child process. */
  while ( (pid = waitpid(-1, &status, WNOHANG)) > 0 ) {
    // find which partition it was on, clear it
    partno = gt::GT_Inst::getInstance().ClearPID(pid);

    // drop back in state to config loaded.  How to reach inside the 
    // statemachine from here to do that?  Need to set up a mutex like we do 
    // for triggering, and put it in the main NGTMasterRC thread running 
    // loop.  We can then force a deconfigure run on the affected partition
    // from within that thread/object

    // Should be careful with I/O in here, this is probably not signal safe.
    // But if we're doing it, let's parse the the status
    if (WIFEXITED(status)) {  // child exited cleanly
      LOG_DEBUG("Progress") << "Cleaned up dead NovaGlobalTrigger on partition " 
                            << partno << " with pid " << pid
                            << " that exited with status " 
			    << WEXITSTATUS(status) << endl;
    }
    if (WIFSIGNALED(status)) {  // child was terminated by a signal
      LOG_DEBUG("Progress") << "Cleaned up dead NovaGlobalTrigger on partition " 
                            << partno << " with pid " << pid
                            << " that was terminated by signal " 
			    << WTERMSIG(status) << endl;
    }
  }
}
