/**
 * The RC comms methods for the NovaGlobalTrigger process.
 * Only deal with run state requests.
 * The Master Controller process does partition making, configuring, etc.
 */

#include <string>
#include <rms/base/RmsDestination.h>
#include <NovaGlobalTrigger/GT.h>
#include <NovaGlobalTrigger/NGTRC.h>
#include <NovaGlobalTrigger/GTGlobals.h>
#include <NovaGlobalTrigger/GT_Stats.h>
#include <DAQDataFormats/TriggerDefines.h>

using namespace gov::fnal::cd::rms;

GT_NAMESPACE_BEGIN

/**
 * Creates an instance with the specified application name.
 */
NGTRC::NGTRC(std::string applicationName) :
  _applicationName(applicationName) {

  // message facility already started my main program, just SetModuleName
  mf::SetModuleName("NGTRC");

  std::vector<std::string> bcastTargets;
  bcastTargets.push_back(base::RmsDestination::ALL_ELEMENT_TARGET);

  _rcReceiver.reset(new rcclient::RunControlReceiver(applicationName, bcastTargets));

  // Use partition number from config globals - NovaGlovalTrigger only
  // exists in one partition at a time

  // Set up the messages this program listens for
  _rcReceiver->addListener<runcontrolmessages::BeginRunRequest,
    NGTRC, runcontrolmessages::BeginRunResponse>
    (partNumber, this);
  _rcReceiver->addListener<runcontrolmessages::StopRunRequest,
    NGTRC, runcontrolmessages::StopRunResponse>
    (partNumber, this);
  _rcReceiver->addListener<runcontrolmessages::PauseRunRequest,
    NGTRC, runcontrolmessages::PauseRunResponse>
    (partNumber, this);
  _rcReceiver->addListener<runcontrolmessages::ResumeRunRequest,
    NGTRC, runcontrolmessages::ResumeRunResponse>
    (partNumber, this);
  _rcReceiver->addListener<runcontrolmessages::StatusRequest,
    NGTRC, runcontrolmessages::StatusResponse>
    (partNumber, this);
  _rcReceiver->addListener<runcontrolmessages::RolloverRunRequest,
    NGTRC, runcontrolmessages::RolloverRunResponse>
    (partNumber, this);

  ::mf::LogInfo("Progress") << "NGTRC thread started" << endl;
}


/**
 * Destructor: clean up anything that we have allocated.
 */
NGTRC::~NGTRC() {
  if (_rcReceiver.get() != 0) {
    _rcReceiver->disconnectFromPartition(partNumber);
    _rcReceiver.reset();
  }
}


/**
 * Thread run method.
 */
void NGTRC::run() {
  _running = true;

  // not much to do - the message handler is taking care of things
  while (! _stopRequested) {
    sleep(5);
  }

  _running = false;
}


/**
 * Handles begin run messages that are received.
 */
boost::shared_ptr<runcontrolmessages::BeginRunResponse> NGTRC::
  handleMessage(boost::shared_ptr<runcontrolmessages::BeginRunRequest> message)
{
  boost::mutex::scoped_lock sl(_callbackMutex);

  // initialize the status that we'll send in the reply to sucess
  rmscore::RequestStatus reqStatus;
  reqStatus.code = 0;
  reqStatus.message = DDS::string_dup("Success");

  // parse the request
  int partitionNumber = message->header.partitionNumber;
  ::mf::LogInfo("State") << "BeginRunRequest received for "
                   << "partition number " << partitionNumber << ".";

  // create the reply
  boost::shared_ptr<runcontrolmessages::BeginRunResponse> transResponse;

  // Change the state
  if ( !changeState(RUNNING) ) {
    ::mf::LogWarning("State") << "Unable to start run" << endl;
    reqStatus.code = 1;
    reqStatus.message = DDS::string_dup("Unable to start run");
    transResponse.reset(new runcontrolmessages::BeginRunResponse());
    transResponse->status = reqStatus;
  } else {
    if (reqStatus.code >= -100) {
      transResponse.reset(new runcontrolmessages::BeginRunResponse());
      transResponse->status = reqStatus;
    }
  }
  return transResponse;
}


/**
 * Handles end run messages that are received.
 */
boost::shared_ptr<runcontrolmessages::StopRunResponse> NGTRC::
  handleMessage(boost::shared_ptr<runcontrolmessages::StopRunRequest> message)
{
  boost::mutex::scoped_lock sl(_callbackMutex);

  // initialize the status that we'll send in the reply to sucess
  rmscore::RequestStatus reqStatus;
  reqStatus.code = 0;
  reqStatus.message = DDS::string_dup("Success");

  // parse the request
  int partitionNumber = message->header.partitionNumber;
  ::mf::LogInfo("State") << "StopRunRequest received for "
                   << "partition number " << partitionNumber << ".";

  // create the reply
  boost::shared_ptr<runcontrolmessages::StopRunResponse> transResponse;

  // Change the state
  if ( !changeState(RUN_CONFIGURED) ) {
    ::mf::LogWarning("State") << "Unable to stop run" << endl;
    reqStatus.code = 1;
    reqStatus.message = DDS::string_dup("Unable to stop run");
    transResponse.reset(new runcontrolmessages::StopRunResponse());
    transResponse->status = reqStatus;
  } else {
    if (reqStatus.code >= -100) {
      transResponse.reset(new runcontrolmessages::StopRunResponse());
      transResponse->status = reqStatus;
      uint64_t counts[NUMBITS];
      transResponse->nEvents = GT_Stats::getInstance().reportTrigNo(counts,NUMBITS);
      // how many triggers had something?
      uint32_t activeTrigs = 0;
      for (uint32_t i=0; i<NUMBITS; i++) 
	if (counts[i] > 0) activeTrigs++;

      // transResponse->evtInfo has dynamic things in there: this sets them up
      transResponse->evtInfo.nEventType = activeTrigs;
      transResponse->evtInfo.evtTypeValue.length(transResponse->evtInfo.nEventType);
      transResponse->evtInfo.evtTypeName.length(transResponse->evtInfo.nEventType);
      transResponse->evtInfo.nEvent.length(transResponse->evtInfo.nEventType);

      // load the sequences.  Loop over trigger IDs, report on non-zero events
      uint32_t n = 0;
      for (uint32_t i=0; i<NUMBITS; i++) {
	if (counts[i] > 0) {
	  transResponse->evtInfo.evtTypeValue[n] = i+1;
	  transResponse->evtInfo.evtTypeName[n] = DDS::string_dup(daqdataformats::getTriggerShortName(GT_Stats::getInstance().Triggers.at(i+1).getType()));
	  transResponse->evtInfo.nEvent[n] = counts[i];
	  n++;
	}
      }
    }
  }
  return transResponse;
}


/**
 * Handles pause run messages that are received.
 */
boost::shared_ptr<runcontrolmessages::PauseRunResponse> NGTRC::
  handleMessage(boost::shared_ptr<runcontrolmessages::PauseRunRequest> message)
{
  boost::mutex::scoped_lock sl(_callbackMutex);

  // initialize the status that we'll send in the reply to sucess
  rmscore::RequestStatus reqStatus;
  reqStatus.code = 0;
  reqStatus.message = DDS::string_dup("Success");

  // parse the request
  int partitionNumber = message->header.partitionNumber;
  ::mf::LogInfo("State") << "PauseRunRequest received for "
                   << "partition number " << partitionNumber << ".";

  // create the reply
  boost::shared_ptr<runcontrolmessages::PauseRunResponse> transResponse;

  // Change the state
  if ( !changeState(RUN_PAUSED) ) {
    ::mf::LogWarning("State") << "Unable to pause run" << endl;
    reqStatus.code = 1;
    reqStatus.message = DDS::string_dup("Unable to pause run");
    transResponse.reset(new runcontrolmessages::PauseRunResponse());
    transResponse->status = reqStatus;
  } else {
    if (reqStatus.code >= -100) {
      transResponse.reset(new runcontrolmessages::PauseRunResponse());
      transResponse->status = reqStatus;
    }
  }
  return transResponse;
}


/**
 * Handles resume run messages that are received.
 */
boost::shared_ptr<runcontrolmessages::ResumeRunResponse> NGTRC::
  handleMessage(boost::shared_ptr<runcontrolmessages::ResumeRunRequest> message)
{
  boost::mutex::scoped_lock sl(_callbackMutex);

  // initialize the status that we'll send in the reply to sucess
  rmscore::RequestStatus reqStatus;
  reqStatus.code = 0;
  reqStatus.message = DDS::string_dup("Success");

  // parse the request
  int partitionNumber = message->header.partitionNumber;
  ::mf::LogInfo("State") << "ResumeRunRequest received for "
                   << "partition number " << partitionNumber << ".";

  // create the reply
  boost::shared_ptr<runcontrolmessages::ResumeRunResponse> transResponse;

  // Change the state
  if ( !changeState(RUNNING) ) {
    ::mf::LogWarning("State") << "Unable to resume run" << endl;
    reqStatus.code = 1;
    reqStatus.message = DDS::string_dup("Unable to resume run");
    transResponse.reset(new runcontrolmessages::ResumeRunResponse());
    transResponse->status = reqStatus;
  } else {
    if (reqStatus.code >= -100) {
      transResponse.reset(new runcontrolmessages::ResumeRunResponse());
      transResponse->status = reqStatus;
    }
  }
  return transResponse;
}


/**
 * Handles heartbeat messages that are received.
 */
boost::shared_ptr<runcontrolmessages::StatusResponse> NGTRC::
  handleMessage(boost::shared_ptr<runcontrolmessages::StatusRequest> message)
{
  boost::mutex::scoped_lock sl(_callbackMutex);

  // initialize the status that we'll send in the reply to sucess
  rmscore::RequestStatus reqStatus;
  reqStatus.code = 0;
  reqStatus.message = DDS::string_dup("Success");

  // parse the request
  //int partitionNumber = message->header.partitionNumber;
  // ::mf::LogInfo("State") << "StatusRequest received for "
  //                 << "partition number " << partitionNumber << ".";

  // create the reply
  boost::shared_ptr<runcontrolmessages::StatusResponse> transResponse;
  if (reqStatus.code >= -100) {
    transResponse.reset(new runcontrolmessages::StatusResponse());
    transResponse->status = reqStatus;
  }
  return transResponse;
}


/**
 * Handles rollover run messages that are received.
 */
boost::shared_ptr<runcontrolmessages::RolloverRunResponse> NGTRC::
  handleMessage(boost::shared_ptr<runcontrolmessages::RolloverRunRequest> message)
{
  boost::mutex::scoped_lock sl(_callbackMutex);

  // initialize the status that we'll send in the reply to sucess
  rmscore::RequestStatus reqStatus;
  reqStatus.code = 0;
  reqStatus.message = DDS::string_dup("Success");

  // parse the request
  int partitionNumber = message->header.partitionNumber;
  ::mf::LogInfo("State") << "RolloverRunRequest received for "
                   << "partition number " << partitionNumber << ".";

  // create the reply
  boost::shared_ptr<runcontrolmessages::RolloverRunResponse> transResponse;

  // reset the counters - does not involve an actual change of state!
  // Can't flub this simple act, so doesn't check for status.  
  // Knock on wood.  I can imagine it flubbing this, but don't know how
  // to check for failure to set a flag.
  if (reqStatus.code >= -100) {
    transResponse.reset(new runcontrolmessages::RolloverRunResponse());
    transResponse->status = reqStatus;
  }

  uint64_t counts[NUMBITS];
  transResponse->nEvents =_rolloverRun(counts,NUMBITS);  
  transResponse->evtInfo.nEventType = 6;
  // transResponse->evtInfo has dynamic things in there: this sets them up
  transResponse->evtInfo.evtTypeValue.length(transResponse->evtInfo.nEventType);
  transResponse->evtInfo.evtTypeName.length(transResponse->evtInfo.nEventType);
  transResponse->evtInfo.nEvent.length(transResponse->evtInfo.nEventType);

  // load the sequences
  // need bits 0->4 and 23
  for (uint32_t i=0; i<5; i++) {
    transResponse->evtInfo.evtTypeValue[i] = i;
// or should this be 2**i?
    transResponse->evtInfo.evtTypeName[i] = DDS::string_dup(daqdataformats::getTriggerShortNameFromMask(0));
    transResponse->evtInfo.nEvent[i] = counts[i];
  }
  transResponse->evtInfo.evtTypeValue[5] = 23;
  transResponse->evtInfo.evtTypeName[5] = DDS::string_dup(daqdataformats::getTriggerShortNameFromMask(23));
  transResponse->evtInfo.nEvent[5] = counts[23];

  return transResponse;
}


/**
 * Begin a run
 */
bool NGTRC::_beginRun()
{
  GT_Stats::getInstance().resetTrigNo();
  { // let triggering happen
      boost::mutex::scoped_lock lock(Triggering_mutex); 
      Triggering = true;
      // notify anyone waiting on this that the state has changed
      Triggering_cond.notify_all();
  }
  GT_Stats::getInstance().startHeartbeat();
  return true;
}


/**
 * Pause a run
 */
bool NGTRC::_pauseRun()
{
  { // Stop triggering
    boost::mutex::scoped_lock lock(Triggering_mutex); 
    Triggering = false;
    // notify anyone waiting on this that the state has changed
    Triggering_cond.notify_all();
  }

  return true;
}


/**
 * Resume a run
 */
void NGTRC::_resumeRun()
{
  { // let triggering happen
      boost::mutex::scoped_lock lock(Triggering_mutex); 
      Triggering = true;
      // notify anyone waiting on this that the state has changed
      Triggering_cond.notify_all();
  }
}


/**
 * Stop a run
 */
void NGTRC::_stopRun()
{
  { // Stop triggering
    boost::mutex::scoped_lock lock(Triggering_mutex); 
    Triggering = false;
    // notify anyone waiting on this that the state has changed
    Triggering_cond.notify_all();
    // tell ganglia we're stopped
    GT_Stats::getInstance().stopHeartbeat();
  }

  ::mf::LogInfo("Progress") << "Run stopping:" << endl;
}


/**
 * rollover a run - just reset counters
 */
uint64_t NGTRC::_rolloverRun(uint64_t counts[], uint32_t numcounts)
{
  ::mf::LogInfo("Progress") << "Run rolling over:" << endl;
  uint64_t TriggerNumber = GT_Stats::getInstance().reportTrigNo(counts,numcounts);
  GT_Stats::getInstance().resetTrigNo();
  return TriggerNumber;
}


GT_NAMESPACE_END
