#include <sstream>
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "NovaGlobalTrigger/GT.h"
#include "NovaGlobalTrigger/NGTStateMachine.h"

using namespace GT_NAMESPACE;

/**
 * Constructor
 */
NGTStateMachine::NGTStateMachine() : _currentState(RUN_CONFIGURED) {

  LOG_DEBUG("NGTStateMachine") << "NGTStateMachine::ctor @" << this << "." << std::endl;
}


/**
 * Destructor
 */
NGTStateMachine::~NGTStateMachine() {

  LOG_DEBUG("NGTStateMachine") << "NGTStateMachine::dtor @" << this << "." << std::endl;
}


/**
 * Static method to return NGTState_t as a string.
 */
std::string NGTStateMachine::NGTStateAsString(NGTState_t state) {

  switch ( state ) {

  case RUN_CONFIGURED:
    return "RUN_CONFIGURED";

  case RUNNING:
    return "RUNNING";

  case RUN_PAUSED:
    return "RUN_PAUSED";

  default:
    ::mf::LogError("NGTStateMachine") << "NGTStateMachine::NGTStateAsString - received unknown state " 
                                << state << "." << std::endl;
  }

  return "UNKNOWN";
}


/**
 * Method to change from _currentState to newState.  If a transition isn't allowed,
 * or if the transition fails, the method returns false.  
 * Transitions must be requested incrementally when moving upward through the state transitions,
 * but larger transition jumps may be requested when moving downward through the
 * state transitions.  (The intermediate downward transition steps will be handled internally.)
 * The reason for this difference in behavior is that you should be able
   to stop a run from anywhere.
 *  
 * @param newState
 * @return true if success, else false.  
 */
bool NGTStateMachine::changeState(NGTState_t newState) {

  ::mf::LogInfo("NGTStateMachine") <<  "NGTStateMachine::changeState - request to move to state "
			     << NGTStateAsString(newState)
			     << " from state " 
			     << NGTStateAsString(_currentState) << std::endl;

  if ( _currentState == RUN_CONFIGURED ) {
    return _changeStateFromRunConfigured(newState);
  }
  else if ( _currentState == RUNNING ) {
    return _changeStateFromRunning(newState);
  }
  else if ( _currentState == RUN_PAUSED ) {
    return _changeStateFromRunPaused(newState);
  }
  
  ::mf::LogError("NGTStateMachine") << "NGTStateMachine::changeState - Unknown current state "
                              << _currentState << "??? Fail." << std::endl;
  return false;
}


/**
 * Private method to change state when current state is RUN_CONFIGURED.
 * @param newState
 * @return true if success, else false.  
 */
bool NGTStateMachine::_changeStateFromRunConfigured(NGTState_t newState) {

  switch ( newState ) {

  case RUN_CONFIGURED:
    return true;

  case RUNNING:
    if ( _beginRun() ) {
      _currentState = newState;
      return true;
    }
    else return false;

  default:
    ::mf::LogError("NGTStateMachine") << "NGTStateMachine::_changeState - request to jump to state " 
                                << NGTStateAsString(newState) << " from state " 
                                << NGTStateAsString(_currentState) << " denied." << std::endl;
    return false;
  }
}


/**
 * Private method to change state when current state is RUNNING.
 * @param newState 
 * @return true if success, else false.  
 */
bool NGTStateMachine::_changeStateFromRunning(NGTState_t newState) {

  switch ( newState ) {

  case RUN_CONFIGURED:
    _stopRun();
    _currentState = newState;
    return true;

  case RUNNING:
    return true;

  case RUN_PAUSED:
    if ( _pauseRun() ) {
      _currentState = newState;
      return true;
    }
    else return false;

  default:
    ::mf::LogError("NGTStateMachine") << "NGTStateMachine::_changeState - request to jump to state " 
                                << NGTStateAsString(newState) << " from state " 
                                << NGTStateAsString(_currentState) << " denied." << std::endl;
    return false;
  }
}


/**
 * Private method to change state when current state is RUN_PAUSED.
 * @param newState 
 * @return true if success, else false.  
 */
bool NGTStateMachine::_changeStateFromRunPaused(NGTState_t newState) {

  switch ( newState ) {

  case RUN_CONFIGURED:
  case RUNNING:
    _resumeRun();
    _currentState = RUNNING;
    return changeState(newState);

  case RUN_PAUSED:
    return true;

  default:
    ::mf::LogError("NGTStateMachine") << "NGTStateMachine::_changeState - request to jump to state " 
                                << NGTStateAsString(newState) << " from state " 
                                << NGTStateAsString(_currentState) << " denied." << std::endl;
    return false;
  }
}
