/*
 * SNEWSSender
 * The application that has to run on SNEWS servers to get us messages
 * 
 * Original 
 * @author Holger Meyer
 * @version $Revision: 1.2 $ $Date: 2014/09/10 15:12:16 $
 *
 * Adapted by
 * J. Zirnstein (janzirn@fnal.gov) for SNEWS Trigger
 *
 */

// System C Libs
#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>
// System C++ Libs
#include <iostream>
#include <string>
#include <fstream>
#include <ctime>
#include <list>
// External Libs
#include <boost/program_options.hpp>
// Custom Libs
#include "XmlRpc/XmlRpc.h"
#include "NovaTimingUtilities/TimingUtilities.h"
#include "NovaDAQUtilities/LocationUtils.h"
#include "NovaSNEWSInterface/SNEWSUtil.h"
#include "NovaSNEWSInterface/SNEWSMessageInfo.h"
#include "NovaSNEWSInterface/SNEWSSenderConfig.h"
extern "C"{
#include "NovaSNEWSInterface/daemon_init.h"
}
// Include the message facility
#include <messagefacility/MessageLogger/MessageLogger.h>

// Manage the Namespaces
namespace po = boost::program_options;

using namespace nova::NovaSNEWSInterface;
using namespace novadaq::nsi;

// Default delay for checking for new SNEWS messages
// the value is in microseconds
#define MESSAGE_POLL_DELAY 1000000

// Structure to hold message details for logging/debugging
struct SNEWSEventData {
  uint32_t evt_type;
  uint32_t evt_no;
  uint64_t time;
};

int main(int argc, char* argv[])
{

  // Initialize Variables
  bool read_events = true;                               // Flag for reading events from the queue

  SNEWSEventData theSNEWSData;                           // Structure to hold the current SNEWS message
  theSNEWSData.evt_type = 0;
  theSNEWSData.time     = 0;
  theSNEWSData.evt_no   = 0;

  int daemon_flag; //flag to run process in daemon mode
  int message_flag; //flag to log output in the message facility

  // Parse command line options using Boost
  std::string cfg_file;

  // Set up absolute location for Server side logging
  std::string absLoc = "/daqlogs/NovaSNEWS/MessageSender/Sender";


  // Declare the supported options.
  po::options_description desc("Allowed options");
  desc.add_options()
    ("help,h", "usage descriptions")
    ("config,c",
     po::value<std::string>(&cfg_file)->
        default_value("srt://NovaSNEWSInterface/config/SNEWSSenderConfig.xml"),
     "The configuration xml file to use")
    ("daemon mode,D", po::value<int>(&daemon_flag)->default_value(0),
     "Flag to run non-interactively in daemon mode. Set to true (1) by default, to disable set 0.")
    ("message facility,M", po::value<int>(&message_flag)->default_value(0),
     "Flag to log output in the message facility. Set to true (1) by default, to disable set 0.");

  po::variables_map vm;
  po::store(po::parse_command_line(argc, argv, desc), vm);
  po::notify(vm);    

  if (vm.count("help")) {
    std::cout << desc << "\n";
    exit(1);
  }

  //daemonizing the application so this can safely run in the background
  if (daemon_flag){
    daemon_init();
  }

  std::string configname = novadaq::LocationUtils::translateURL(cfg_file);

  //Pull in the XML config file
  std::string xmlString;
  {
    std::ifstream inStream(configname.c_str());
    if (! inStream.is_open()) {
      std::cerr << "Configuration | FileNotFound \n"
		<< "Unable to open " << configname << "." << std::endl;
      exit(-1);
    }
    else {
      std::string line;
      while (std::getline(inStream, line)) {
	xmlString.append(line);
      }
      inStream.close();
    }
  }
 
  //connect objects in XSD file to settings in XML file
  novadaq::nsi::SNEWSSenderConfigType SSCfg;
  try {
    SSCfg.deserialize(xmlString);
  }
  catch (...) {
    std::cerr << "Configuration | Deserialization failed\n"
              << "Syntax error in your configuration xml file?\n"
              << " File content (" << configname << ") :\n"
              << xmlString << "\n Exception "
              << std::endl;
    exit (-1);
  }

  // Set up log file
  std::string logFname;
  {
    std::stringstream fname;
    timeval now;
    struct tm* ptm;
    char timeStr[40];
    gettimeofday(&now,0);
    ptm = localtime (&now.tv_sec);
    strftime (timeStr, sizeof (timeStr), "%Y-%m-%d_%H-%M-%S", ptm);
    fname<< absLoc <<"_"<< timeStr <<"_log.txt";
    logFname=fname.str();
  }

  // Start the Message Facility
  if(message_flag){
    mf::StartMessageFacility(mf::MessageFacilityService::MultiThread,
//			     mf::MessageFacilityService::logFile(logFname));
			     mf::MessageFacilityService::logConsole());
    mf::SetApplicationName("SNEWSSender");
    mf::SetModuleName("Main");
  }

  // Get Destination information
  novadaq::nsi::SNEWSMessageDestinationType const& destCfg = 
    SSCfg.getSNEWSMessageDestination();

  std::string rpcHost = destCfg.getSNEWSDestination().getHost();
  int rpcPort = destCfg.getSNEWSDestination().getPort();

  // Be vocal about what you're sending where
  std::cout << "Sending spill information to " << rpcHost << ":"
            << rpcPort << std::endl;
  ::mf::LogInfo("Initialize")<< "Sending spill information to " << rpcHost << ":"
                             << rpcPort;

  SNEWSMessageTypeSetType const& smts = destCfg.getSNEWSMessageTypeSet();

  //get list of message types from configuration
  std::list<int> smtypes;
  for (int j=0; j!= smts.getSNEWSMessageTypeCount(); ++j) {
    SNEWSMessageTypeType st = smts.getSNEWSMessageType(j);
    int type = SNEWSMessageInfo::getMessageTypeFromName(std::string(st.getName()));
    if (find(smtypes.begin(),smtypes.end(),type) == smtypes.end()
        && type>-1) {
      smtypes.push_back(type);
    }
  }

  // What messages are we capable of sending?
  std::cout << "SNEWS Messages to generate: ";
  for (std::list<int>::const_iterator it = smtypes.begin();
       it != smtypes.end();   ++it)
  {
    std::cout << SNEWSMessageInfo::getMessageNameFromType((*it)) << " ";
    ::mf::LogInfo("Message")<<"Registering message type to trigger: "<<SNEWSMessageInfo::getMessageNameFromType((*it));
  }
  std::cout << std::endl;

  // set up XmlRpcClient
  XmlRpc::setVerbosity(1);
  XmlRpc::XmlRpcClient c(rpcHost.c_str(),
                         rpcPort);
  XmlRpc::XmlRpcValue result;

  std::string str;
  MessageType messageType;
  XmlRpc::XmlRpcValue messageInfoXml;


  std::cout << "Sending messages..." << std::endl;
  ::mf::LogInfo("Messages")<<"Sending Messages...";

  bool send = true;
  int msgSent = 0;

  // Main event read loop
  // We stay in this loop until told to exit it
  while (read_events) {

    //Dumb test for sender
    uint64_t novaTime;
    {
      timeval trigNow;
      struct tm* ttm;
      gettimeofday(&trigNow,0);
      ttm = localtime (&trigNow.tv_sec);
      novadaq::timeutils::convertUnixTimeToNovaTime(*ttm, novaTime);
    }

    // Load the information into a structure
    theSNEWSData.evt_type = messageType;
    theSNEWSData.time     = novaTime;
    theSNEWSData.evt_no++;

    // decode time into messageInfoXml
    SNEWSUtil::gpstimeToString(novaTime, str);
    messageInfoXml[0] = str;

    //decode type into messageInfoXml
    messageType = kSNEWSBeatFast;
    messageInfoXml[1] = messageType;
    if (send) {
      if (c.execute("SNEWSMessage", messageInfoXml, result)) {
        std::cout << "Message method result: " << result << "\n";
        std::cout << "Type " 
                  << SNEWSMessageInfo::getMessageNameFromType(messageType)
                  <<"("<<messageType<<")"
                  << novaTime << std::endl;
        ::mf::LogInfo("MessageSent")<<"Messagetype: "
                                  <<SNEWSMessageInfo::getMessageNameFromType(messageType)
                                  <<"("<<messageType<<")"
                  		  << novaTime << std::endl;
      }
      else {
        std::cout << "SNEWSSender: Error sending 'Message'\n";
        ::mf::LogError("Message")<<"SNEWSSender: Error sending 'Message'";
      }
    } // endif send

/*    
    std::cout << "Message method result: " << result << "\n";
    std::cout << "Type " 
              << SNEWSMessageInfo::getMessageNameFromType(messageType)
              << "\nTime "
              << novaTime << std::endl;
*/

    msgSent++;
    if (msgSent > 10) read_events = false; 
    usleep(MESSAGE_POLL_DELAY); // Pause before sending next message

  } // endwhile
 
  return 0;
}
