/**
 * NssSpillReceiver main file
 * $Id: NssSpillReceiver.cc,v 1.15 2015/03/02 18:31:55 janzirn Exp $
 * Receive information on spills through XmlRpc and notify GT through rms.
 * This will eventually run at the far detector.
 */
// System Libs
#include <iostream>
#include <cstdlib>
// External Libs
#include <boost/program_options.hpp>
// Custom Libs
#include "NovaSpillServer/NssXmlRpcServerClasses.h"
#include "rms/provider/DDSConnection.h"
#include "rms/base/RmsDestination.h"
#include "rms/RmsReceiver.h"
#include "rms/RmsSender.h"

using namespace XmlRpc;
using namespace gov::fnal::cd::rms;
using namespace nova::NovaSpillServer;
namespace po = boost::program_options;

#ifndef TRUE
#define TRUE   1
#endif /* TRUE */

#ifndef FALSE
#define FALSE  0
#endif /* FALSE */

#define CLIENT_APP_NAME "NssSpillReceiver"
#define TRIGGER_MAILBOX "SpillServerMailbox"

// Include the message facility
#include <messagefacility/MessageLogger/MessageLogger.h>

// The server
XmlRpcServer s;

int main(int argc, char* argv[])
{

  int port=-1;
  int sourceport = -1;
  int daq_partition;
  int rms_partition;
  int message_flag = TRUE;

  // Declare the supported command line options.
  po::options_description desc("Allowed options");
  desc.add_options()
    ("help,h", "usage descriptions")
    ("port,p", po::value<int>(&port)->default_value(-1),
     "The port number to receive xmlrpc spill messages on, must be specified.")
    ("rms_partition,r",
     po::value<int>(&rms_partition)->
     default_value(base::RmsDestination::NULL_PARTITION),
     "The DDS partition number to send received spill information to. (Run multiple receivers to send spills to multiple partitions.)")
    ("daq_partition,z",
     po::value<int>(&daq_partition)->
     default_value(0),
     "The DAQ partition number associated with this receiver. (Run one receivers per DAQ partition.)")
    ("source-port,s", po::value<int>(&sourceport)->default_value(-1),
     "The port number of a spill-forwarder on localhost to request spills from.")
    ("message facility,M", po::value<int>(&message_flag)->default_value(1),
     "Flag to log output to the message facility. Set to true (1) by default, to disable set 0.")
    ;
  po::positional_options_description p;
  p.add("port", 1);
  p.add("rms_partition", 2);

  std::stringstream fname;
  std::string logFname;
  timeval now;
  struct tm* ptm;
  char timeStr[40];
  gettimeofday(&now,0);
  ptm = localtime (&now.tv_sec);
  strftime (timeStr, sizeof (timeStr), "%Y-%m-%d_%H-%M-%S", ptm);
  fname<<"NssSpillReceiver_partition"<<rms_partition<<"_"<<port<<"_"<<timeStr<<"_log.txt";
  logFname=fname.str();

  po::variables_map vm;
  po::store(po::command_line_parser(argc, argv).
	    options(desc).positional(p).run(), vm);
  po::notify(vm);

  if (vm.count("help") || (port == -1)) {
    cout << desc << "\n";
    return 1;
  }

  // Start the Message Facility
  if (message_flag){
    mf::StartMessageFacility(mf::MessageFacilityService::MultiThread,
  			     mf::MessageFacilityService::logConsole());
    mf::SetApplicationName("NssSpillReceiver");
    mf::SetModuleName("Main");
  }

  //Connect to RMS, get ready to send a message to rms for each spill
  // create the connection to RMS
  boost::shared_ptr<provider::DDSConnection>
    rmsConnection(new provider::DDSConnection(CLIENT_APP_NAME, rms_partition));

  // create the destination to represent where the messages will be sent
  base::RmsDestination sendDest(TRIGGER_MAILBOX,
                       base::RmsDestination::TRIGGER_CHANNEL);

  // create a sender with the connection and destination
  boost::shared_ptr< RmsSender<provider::DDSConnection,
                     nssmessages::SpillMessage> >
    requestSenderptr(new RmsSender<provider::DDSConnection,
                                   nssmessages::SpillMessage>
                           (rmsConnection, sendDest));

  //XmlRpc Methods:
  Spill spill(&s, requestSenderptr);

  XmlRpc::setVerbosity(0);

  // Create the server socket on the specified port
  s.bindAndListen(port);
  //std::cout<<"NssSpillReceiver setup to listen to spills sent to port: "<<port<<" on partition: "<<partition<<std::endl;
  ::mf::LogInfo("Initialize")<<"NssSpillReceiver setup to listen to spills sent to port: "
			     <<port<<" on DDS partition: "<<rms_partition<<" associated with DAQ partition: "<<daq_partition;

  // Send request to receive Spills
  if (sourceport>1024 && sourceport <=65535 && sourceport != port) {
    // See if a forwarder is listening on sourceport.
    // If so, ask for spills to be sent to localhost:port, i.e. this process
      XmlRpc::setVerbosity(0);

      XmlRpcClient c("localhost",sourceport);
      XmlRpcValue result, params;
      params[0] = "localhost";
      params[1] = port;
      //std::cout << "Sending DestinationAdd(" << "localhost" << ":"
	//      << port << ")..." << std::endl;
      ::mf::LogInfo("Initialize")<< "Sending DestinationAdd(" << "localhost" << ":"
	                         << port << ")...";
      c.execute("DestinationAdd",params,result);
      if (std::string(result).size()!=0) {
	;;std::cout << "DestinationAdd result: " << result <<std::endl;
        ::mf::LogInfo("Initialize")<< "DestinationAdd result: " << (std::string)result;
      } else {
	//std::cout << "Registering with a SpillForwarder failed. I will "
	//	  << "continue to listen for spills\n" << "on port " << port 
	//	  << ". You have to make sure someone is going to send them."
	//	  << std::endl;
        ::mf::LogError("RegisterFail")<< "Registering with a SpillForwarder failed. I will "
		                      << "continue to listen for spills\n" << "on port " << port 
		                      << ". You have to make sure someone is going to send them.";
      }
  }

  // Wait for requests indefinitely
  s.work(-1.0);

  requestSenderptr->close();
  rmsConnection->close();
  return 0;
}

