/**
 * NssSpillReceiverEavesDropper
 * $Id: NssSpillReceiverEavesDropper.cc,v 1.11 2012/12/23 22:10:56 hmeyer Exp $
 * Listen to RMS messages issued by NssSpillReceiver and print
 * to stdout. Also warn if no messages are received for a while.
 */
// System Libs
#include <ctime>
#include <iostream>
#include <sstream>
#include <fstream>
// External Libs
#include <boost/program_options.hpp>
#include <boost/shared_ptr.hpp>
// Custom Libs
#include <rms/provider/CETDDS.h>
#include <rms/provider/DDSConnection.h>
#include <rms/base/RmsDestination.h>
#include <rms/RmsReceiver.h>
#include <rms/RmsSender.h>

#include <DAQMessages/ccpp_NssMessages.h>
#include <NovaTimingUtilities/TimingUtilities.h>
#include <NovaSpillServer/NssSpillInfo.h>

using namespace gov::fnal::cd::rms;
using namespace nova::NovaSpillServer;
namespace po = boost::program_options;

#define CLIENT_APP_NAME "NssSpillReceiverEavesDropper"
#define TRIGGER_MAILBOX "SpillServerMailbox"

class SpillListener {
 public:
  SpillListener(boost::shared_ptr<provider::DDSConnection> ddsConnection);
  void processMessage(boost::shared_ptr<nssmessages::SpillMessage> spillMessage);
  int getMessageCount() {return _msgCount;}
  void resetMessageCount() {_msgCount = 0;}
  void setLogFile(std::ofstream* ptr) {_logFile = ptr;}
 private:
  boost::shared_ptr<provider::DDSConnection> _ddsConnection;
  int _msgCount;
  ofstream* _logFile;
};

SpillListener::
SpillListener(boost::shared_ptr<provider::DDSConnection> ddsConnection) {
  _ddsConnection = ddsConnection;
  _msgCount = 0;
  _logFile = 0;
}

void SpillListener::
processMessage(boost::shared_ptr<nssmessages::SpillMessage> spillMessage) {

  // copy the message data to a spillinfo object
  NssSpillInfo spillinfo(spillMessage->time,spillMessage->type);
  
  //compute difference from unix system time
  timeval now;
  uint64_t unixNow, deltaNowFromSpill;
  gettimeofday(&now,0);
  novadaq::timeutils::convertUnixTimeToNovaTime(now,unixNow);
  deltaNowFromSpill = unixNow - spillinfo.getTime();

  // report the spill time
  std::cout << "Received "
	    <<  NssSpillInfo::getSpillNameFromType(spillinfo.getType())
	    << " spill at time " << std::dec << spillinfo.getTime()
	    << " = 0x" << std::hex << spillinfo.getTime()
	    << std::dec << "with delay " << deltaNowFromSpill << "."
	    << std::endl;

  // increment count
  ++_msgCount;

  // log time and delay for later analysis
  if (_logFile) {
    (*_logFile) << spillinfo.getType() << " " << std::hex 
		<< spillinfo.getTime() << " "
		<< deltaNowFromSpill << std::endl;
  }
}

int main(int argc, char* argv[]) {
  int partition, logging;
  std::string logFname;
  ofstream logFile;
  
  // Declare the supported command line options.
  po::options_description desc("Allowed options");
  desc.add_options()
    ("help,h", "usage descriptions")
    ("partition,p",
     po::value<int>(&partition)->
     default_value(base::RmsDestination::NULL_PARTITION),
     "The DDS partition number to listen for spill information.")
    ("log,l", po::value<int>(&logging)->default_value(0),
     "Logging level: 0 none, 1 on")
    ("filename,f", po::value<std::string>(&logFname),
     "Filename for logfile if given, default used otherwise")
    ;
  po::positional_options_description p;
  p.add("partition", -1);
  
  po::variables_map vm;
  po::store(po::command_line_parser(argc, argv).
	    options(desc).positional(p).run(), vm);
  po::notify(vm);
  
  if (vm.count("help")) {
    cout << desc << "\n";
    return 1;
  }
  
  // create the connection to RMS
  boost::shared_ptr<provider::DDSConnection>
    rmsConnection(new provider::DDSConnection(CLIENT_APP_NAME,
					      partition));
  
  // create the destination that represents where we will receive
  // messages from
  base::RmsDestination receiveDest(TRIGGER_MAILBOX,
				   base::RmsDestination::TRIGGER_CHANNEL);
  
  // create a receiver with the connection object and destination
  RmsReceiver<provider::DDSConnection,
    nssmessages::SpillMessage,
    SpillListener>
    requestReceiver(rmsConnection, receiveDest);
  
  SpillListener* spillListener = new SpillListener(rmsConnection);
  boost::shared_ptr<SpillListener> theListener(spillListener);
  requestReceiver.setListener(theListener);
  
  if (logging) {
    //open file, set up tree
    if (logFname=="") {
      std::stringstream fname;
      timeval now;
      struct tm* ptm;
      char timeStr[40];
      gettimeofday(&now,0);
      ptm = localtime (&now.tv_sec);
      strftime (timeStr, sizeof (timeStr), "%Y-%m-%d_%H-%M-%S", ptm);
      fname<<"SpillLog_partition"<<partition<<"_"<<timeStr<<".txt";
      logFname=fname.str();
    } else {
      //trust that the user specified a valid file name. Should check here.
    }
    logFile.open(logFname.c_str());
    logFile << "NovaSpillServerEavesDropper Spill log: Type, NovaTime, DeltaTnowFromSpill"<<std::endl;
    spillListener->setLogFile(&logFile);
  }
  
  std::cout << "Waiting for spill messages..." << std::endl;
  
  while (true) {
    sleep(60);
    std::cout << "Number of spills in the last minute: "
	      << theListener->getMessageCount()
	      << std::endl;
    theListener->resetMessageCount();
  }
  
  // cleanup and exit
  requestReceiver.close();
  rmsConnection->close();
  if (logging) {
    logFile.close();
  }
  return 0;
}
