/**
 * NssTDUApp
 * $Id: NssTDUApp.cc,v 1.27 2014/11/05 16:18:14 edniner Exp $
 * The application that has to run on the Master TDU to serve spills
 * 
 * @author Holger Meyer
 * @version $Revision: 1.27 $ $Date: 2014/11/05 16:18:14 $
 */
// System C Libs
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <setjmp.h>
// System C++ Libs
#include <iostream>
#include <string>
#include <fstream>
#include <ctime>
#include <list>
// External Libs
#include <boost/program_options.hpp>
// Custom Libs
#include "XmlRpc/XmlRpc.h"
#include "TDUUtilities/tdu.h"
#include "NovaTimingUtilities/TimingUtilities.h"
#include "NovaDAQUtilities/LocationUtils.h"
#include "NovaSpillServer/NssSpillInfo.h"
#include "NovaSpillServer/NssUtil.h"
#include "NovaSpillServer/NssTDUAppConf.h"
#include "NovaSpillServer/NssTDUAppConfig.h"
extern "C"{
#include "NovaSpillServer/daemon_init.h"
}
// Include the message facility
#include <messagefacility/MessageLogger/MessageLogger.h>

// Manage the Namespaces
namespace po = boost::program_options;

using namespace nova::NovaSpillServer;
using namespace novadaq::nss;

// Default delay for checking for new beam events
// the value is in microseconds
#define SPILL_POLL_DELAY 1000000

// Define Macros for managing the Shared Memory Segment
#define SHM_MODE (0666)
#define SHM_MODE_CREAT (IPC_CREAT | 0666)

#define SHM_SIZE 0x80000  /* 512k */

#ifndef TRUE
#define TRUE   1
#endif /* TRUE */

#ifndef FALSE
#define FALSE  0
#endif /* FALSE */

#ifndef FAIL
#define FAIL  -1
#endif /* FAIL */

// Global Constants
static uint64_t const fifteensecs = 15*64000000;

// Global Variables
uint64_t lastspill_time;
uint64_t lastTCR_time;
uint64_t tcrCount;

// global termination flag
bool termination_pending = false;

// Jump Points for program termination
static sigjmp_buf termbuffer;          // Jump point for termination
static volatile sig_atomic_t canjump;  

static void sigterm(int signo){
  printf("Received request for termination");
  termination_pending = true;
  // siglongjmp(termbuff, 1);
}

void setup_unix_signal_handlers(){
  struct sigaction term;
  
  term.sa_handler = sigterm;
  sigemptyset(&term.sa_mask);
  term.sa_flags = 0;
  term.sa_flags |= SA_RESTART;
  sigaction(SIGTERM, &term, 0);
}


struct AccelEventData {
  uint32_t evt_type;
  uint32_t evt_no;
  uint64_t time;
};

struct SHMSpillDataMap {
  uint32_t* data_start;
  uint32_t* data_end;
  struct AccelEventData* data_insert;
  uint32_t data_entries;
};

int main(int argc, char* argv[])
{

  // Initialize Variables
  lastspill_time = 0;
  lastTCR_time = 0;
  tcrCount = 0;

  // int    memseg_flag    = FALSE;                         // Flag indicating we set a mem segement id
  int    memid          = -1;                            // Base ID for the shm segment (0x0 = IPC_PRIVATE)
  unsigned int basesize = 0x0000400;                     // Base size for the initial shm query (1k)
  unsigned int memsize  = 0x2000000;                     // Default size of shared memory segment is 32M
  int    shmid          = 0;
  char*  shmptr         = NULL;                          // Pointer to the start of the shared memory buffer
  // int*   shmptr_ins     = NULL;                          // Offset for insertion into the SHM buffer
  int*   shmptr_int     = NULL;                          // Pointer (integer word) to the start of the shared memory buffer
  char*  shmptr_start   = NULL;                          // Pointer to the start of the data area in the shared memory buffer
  char*  shmptr_end     = NULL;                          // Pointer to the end of the data area in the shared memory buffer
  char*  shmptr_current = NULL;                          // Pointer to the current insertion point in the data buffer

  bool read_events = true;                               // Flag for reading events from the queue


  struct SHMSpillDataMap* DataMap;                       // Pointer to the Shared Memory Data Map
  struct AccelEventData*  AccelData;                     // Pointer to the start of the Accelerator Data block

  AccelEventData theAccelData;                           // Structure to hold the current accelerator event
  theAccelData.evt_type = 0;
  theAccelData.time     = 0;
  theAccelData.evt_no   = 0;

  int daemon_flag = TRUE; //flag to run process in daemon mode
  int message_flag = TRUE; //flag to log output in the message facility
  int server_flag = TRUE; //flag to log output to server in addition to console

  bool BaseClock128 = false;

  // Parse command line options using Boost
  std::string cfg_file;

  // Declare the supported options.
  po::options_description desc("Allowed options");
  desc.add_options()
    ("help,h", "usage descriptions")
    ("config,c",
     po::value<std::string>(&cfg_file)->
        default_value("srt://NovaSpillServer/config/NssTDUAppConfig.xml"),
     "The configuration xml file to use")
    ("daemon mode,D", po::value<int>(&daemon_flag)->default_value(1),
     "Flag to run non-interactively in daemon mode. Set to true (1) by default, to disable set 0.")
    ("message facility,M", po::value<int>(&message_flag)->default_value(1),
     "Flag to log output in the message facility. Set to true (1) by default, to disable set 0.")
    ("server mode,S", po::value<int>(&server_flag),
     "Flag to log messages to server (requires running DDS) in addition to console.  Set to true (1) be default, to disable set 0.")
#if 0
    ("shm,s",
     po::value<int>(&memid)->default_value(0x0),
     "Memory Segment identifier to connect to (hex)"),
    ("mem,m",
     po::value<int>(&memid)->default_value(0x0),
     "Memory Segment identifier to connect to (4 char)")
#endif
    ;

  po::variables_map vm;
  po::store(po::parse_command_line(argc, argv, desc), vm);
  po::notify(vm);    

  if (vm.count("help")) {
    std::cout << desc << "\n";
    exit(1);
  }

  //daemonizing the application so this can safely run in the background
  if (daemon_flag){
    daemon_init();
  }

  // Retrieve the hostname of the TDU we are running on
  char hostname[1024];
  size_t len=1024;
  gethostname(hostname, len);

  std::string configname = novadaq::LocationUtils::translateURL(cfg_file);

  //Pull in the XML config file
  
  std::string xmlString;
  {
    std::ifstream inStream(configname.c_str());
    if (! inStream.is_open()) {
      // ::mf::LogError("Configuration | FileNotFound") 
      std::cerr << "Configuration | FileNotFound \n"
		<< "Unable to open " << configname << "." << std::endl;
      exit(-1);
    }
    else {
      std::string line;
      while (std::getline(inStream, line)) {
	xmlString.append(line);
      }
      inStream.close();
    }
  }
 
  //connect objects in XSD file to settings in XML file
  novadaq::nss::NssTDUAppConfigType NssTACfg;
  try {
    NssTACfg.deserialize(xmlString);
  }
  //catch (novadaq::cfg::xsd::cxx::tree::parsing<char>& excpt) {
  catch (...) {
    std::cerr << "Configuration | Deserialization failed\n"
              << "Syntax error in your configuration xml file?\n"
              << " File content (" << configname << ") :\n"
              << xmlString << "\n Exception "
//            << excpt 
              << std::endl;
    exit (-1);
  }

  std::stringstream fname;
  std::string logFname;
  timeval now;
  struct tm* ptm;
  char timeStr[40];
  gettimeofday(&now,0);
  ptm = localtime (&now.tv_sec);
  strftime (timeStr, sizeof (timeStr), "%Y-%m-%d_%H-%M-%S", ptm);
  fname<<"NssTDUApp"<<"_"<<timeStr<<"_log.txt";
  logFname=fname.str();
  // Start the Message Facility
  if(message_flag){
    if (server_flag){
      mf::StartMessageFacility(mf::MessageFacilityService::MultiThread,
			       mf::MessageFacilityService::logCS());
    }
    else{
      mf::StartMessageFacility(mf::MessageFacilityService::MultiThread,
			       mf::MessageFacilityService::logConsole());
    }
    mf::SetApplicationName("NssTDUApp");
    mf::SetModuleName("Main");
  }
  
  novadaq::nss::TDUSpillDestinationType const& destCfg = 
    NssTACfg.getTDUSpillDestination();

  std::string rpcHost = destCfg.getTDUDestination().getHost();
  int rpcPort = destCfg.getTDUDestination().getPort();

  std::cout << "Sending spill information to " << rpcHost << ":"
            << rpcPort << std::endl;
  ::mf::LogInfo("Initialize")<< "Sending spill information to " << rpcHost << ":"
                             << rpcPort;

  TDUSpillTypeSetType const& tsts = destCfg.getTDUSpillTypeSet();

  //get list of spilltypes from configuration
  std::list<int> sptypes;
  for (int j=0; j!= tsts.getTDUSpillTypeCount(); ++j) {
    TDUSpillTypeType st = tsts.getTDUSpillType(j);
    int type = NssSpillInfo::getSpillTypeFromName(std::string(st.getName()));
    if (find(sptypes.begin(),sptypes.end(),type) == sptypes.end()
        && type>-1) {
      sptypes.push_back(type);
    }
  }

  //Now take more detailed configuration options into account
  //and put the event mask together
  novadaq::nss::TDUAppConfType const& tappCfg = NssTACfg.getTDUAppConf();

  //see description of TDU registers in nova-doc-5370.
  //Should get these from some header file:
  uint32_t evtMask = 0x0; //set below, write to reg. 0x23
  uint32_t ctrlMask = 0x0400; //bit 10: enable events, write to reg. 0x0
  uint32_t irqMask = 0x0016; //bit 4: enable PowerPC interrupt, to reg 0x24

  uint32_t enableA9   = (1<<0);
  uint32_t enableAD   = (1<<1);
  uint32_t enable1F   = (1<<2);
  uint32_t enable1D   = (1<<3);
  uint32_t enable8f   = (1<<4);
  uint32_t enable00   = (1<<5);
  uint32_t enableA4   = (1<<6);
  uint32_t enableA5   = (1<<7);
  uint32_t enableNuMI = (1<<8);
  uint32_t enableBNB  = (1<<12);

  // 1Hz Accel Pulser
  evtMask |= enable8f;              // Enable these regardless of config files
                                    // since they are used for sanity checks

  // Alternatively this will only enable 1Hz spills if they are specified
  // in the config files
  if (find(sptypes.begin(),sptypes.end(),kAccelOneHztclk) != sptypes.end()){
    evtMask |= enable8f;
  }

  //NuMI BSYNC Triggers Enabled by the config file (MIBS-74)
  if (find(sptypes.begin(),sptypes.end(),kNuMI) != sptypes.end()) {
    evtMask |= enableNuMI;
  }

  // Booster Neutrino Beam (BNB) Triggers Enabled by the config file
  if (find(sptypes.begin(),sptypes.end(),kBNB) != sptypes.end()) {
    evtMask |= enableBNB;
  }

  // NuMI TCLK Triggers Enabled by the config file ($A9 and $AD)
  if (find(sptypes.begin(),sptypes.end(),kNuMItclk) != sptypes.end()) {
    if (tappCfg.NuMItclkIssueOnA9()) {
      evtMask |= enableA9;
    }
    if (tappCfg.NuMItclkIssueOnAD()) {
      evtMask |= enableAD;
    }
    if ((!tappCfg.NuMItclkIssueOnA9()) && (!tappCfg.NuMItclkIssueOnAD())) {
      std::cout << "Warning: configuration asks for NuMItclk spills,"
                << " but has both $AD and $A9 disabled. Choose one!"
                << std::endl;
      ::mf::LogWarning("NuMItclk")<< "Warning: configuration asks for NuMItclk spills,"
                                  << " but has both $AD and $A9 disabled. Choose one!";
    }
  }
  // Booster Neutrino Beam (BNB) Triggers Enabled by the config file ($1F extraction kicker)
  if (find(sptypes.begin(),sptypes.end(),kBNBtclk) != sptypes.end()) {
    evtMask |= enable1F;
    if (tappCfg.BNBtclkuse1Dand1F()) evtMask |= enable1D;
  }
  //SuperCycle
  if (find(sptypes.begin(),sptypes.end(),kSuperCycle) != sptypes.end()) {
    evtMask |= enable00;
  }
  //NuMISampleTrig
  if (find(sptypes.begin(),sptypes.end(),kNuMISampleTrig) != sptypes.end()) {
    evtMask |= enableA4;
  }
  //NuMIReset
  if (find(sptypes.begin(),sptypes.end(),kNuMIReset) != sptypes.end()) {
    evtMask |= enableA5;
  }

  std::cout << "Spills to generate: ";
  for (std::list<int>::const_iterator it = sptypes.begin();
       it != sptypes.end();   ++it)
  {
    std::cout << NssSpillInfo::getSpillNameFromType((*it)) << " ";
    ::mf::LogInfo("Spill")<<"Registering spill type to trigger: "<<NssSpillInfo::getSpillNameFromType((*it));
  }
  std::cout << std::endl;

  // get some configuration
  NssTDUAppConf cnf;

  // set up XmlRpcClient
  XmlRpc::setVerbosity(0);
  XmlRpc::XmlRpcClient c(rpcHost.c_str(),
                         rpcPort);
  XmlRpc::XmlRpcValue result;

  // set up the TDU to send spills based on configuration
  nova::TDUUtilities::tdu mtu;

  // Readout the TDU's Firmware Version
  uint32_t FPGAFirmwareVersion = mtu.readFPGAFirmwareVersion();

  // Peform Firmwave version specific configurations

  //bool subfifteen = (FPGAFirmwareVersion <= 0x0106);
  bool subfifteen = false;

  if(FPGAFirmwareVersion <= 0x0106){
    // For a while the TDU firmware converted GPS time into nova-time using
    // a nova-epoch that was wrong by 15 seconds. (# of leap seconds)
    // This is fixed with the newer firmware versions.
    //std::cout << "Firmware version 0x" << std::hex << FPGAFirmwareVersion
    //          << std::dec << " detected." << std::endl;
    subfifteen = true;
  }

  // As of version 2.10 the base clock for the tdu is 128MHz
  if(FPGAFirmwareVersion >= 0x020a){
    BaseClock128 = true;
  }

  // Setup the control Registers
  uint32_t ctrl = mtu.getRegister(cnf.ctrlReg);
  ctrl |= ctrlMask;
  mtu.setRegister(cnf.ctrlReg,ctrl);

  // Setup the accelerator event register
  mtu.setRegister(cnf.evtReg,evtMask);

  // Setup the interupt register
  uint32_t irq = mtu.getRegister(cnf.irqReg);
  irq |= irqMask;
  mtu.setRegister(cnf.irqReg,irq);

  std::string str;
  SpillType spillType;
  XmlRpc::XmlRpcValue spillInfoXml;

  // start to send out spills
  uint16_t buffer[4096];


  /****************************************************
   * Setup a shared memory segment to give us a small *
   * history/spy buffer that we can look at to parse  *
   * out beam spill times                             *
   ****************************************************/
  char seg[]="BEAM\0";
  memcpy(&memid,seg,4);

  printf("%d memid\n",memid);

  if(memid != -1){
    // IPC_PRIVATE Memory Segement
    if(memid == 0x0){
      printf("No memory segment identifier set\nAllocating Private Segment");
      ::mf::LogWarning("MemSegment")<<"No memory segment identifier set, Allocating private segment";
      shmid = shmget(IPC_PRIVATE, memsize, SHM_MODE);
      if(shmid){
	shmptr = (char*) shmat(shmid, 0,0);
      }else{
	printf("Failed to attach or create a PRIVATE shared memory segment");
        ::mf::LogError("MemSegment")<<"Failed to attach or create a PRIVATE shared memory segment, exiting";
	exit(FAIL);
      } // endif shmid
    } // endif IPC_PRIVATE


    // Public Named Segment
    
    if(memid != 0x0){
      // First Check to see if the segment already exists and if it does get 
      // the stats on it so we can check it's current size
      struct shmid_ds shm_seg_info;

      printf("Attempting to connect to shared memory\nID: 0x%08x\n",memid);
      ::mf::LogInfo("MemSegment")<<"Attempting to connect to shared memory ID: 0x"<<std::hex<<memid;
      shmid = shmget(memid, basesize, SHM_MODE);

      if(shmid >= 0){
	if(shmid == 0){
	  // Sigh we should never have a zero ID for the segment, but on the old TDU linux we 
	  // do get a zero id the very first time we try to make a segment after a reboot
	  printf("Received shared memory segment with ID = 0.  This is not standard, but tolerated\n");
          ::mf::LogWarning("ShmId0")<<"Received shared memory segment with ID = 0. This is not standard, but tolerated";
	}
	// Check that we got an existing segment
	// Retrieve the information on the segment (mainly the size)
	int result = shmctl(shmid, IPC_STAT, &shm_seg_info);
	if(result < 0){
	  perror("Unable to start shared memory segment"); 
          ::mf::LogError("shmFail")<<"Unable to start shared memory segment";
	  exit(-1);
	}
	// Check the existing size of the segment and compare it against
	// the requested size
	if( shm_seg_info.shm_segsz < memsize){
	  // The existing segment has a size that is less than the requested size
	  printf("Existing Shared Memory segment has a size less than the requested minimum (Requested: %u, Actual: %zu\n)\n",
		 memsize,shm_seg_info.shm_segsz);
          ::mf::LogWarning("MemSegment")<<"Existing shared memory segment has a size less then the requested minimum, Requested: "
                                        <<memsize<<", Actual: "<<shm_seg_info.shm_segsz;
	};
	shmptr = (char*) shmat(shmid, 0,0);
      }else{
	printf("Failed to attach to existing shared memory segment using:\nID: %d\tSize:%d\n",memid,memsize);
	printf("Creating new segment with Identifier: 0x%08x, Size: %u bytes\n",memid,memsize);
        ::mf::LogWarning("MemSegment")<<"Failed to attach to existing shared memory segment using ID: "
                                      <<memid<<" Size: "<<memsize<<" Creating new segment with Identifier: 0x"
                                      <<std::hex<<memid<<std::dec<<", Size: "<<memsize<<" bytes";
	shmid = shmget(memid, memsize, SHM_MODE_CREAT);
	if(shmid){
	  shmptr = (char*) shmat(shmid, 0,0);
	}else{
	  printf("Failed to attach or create a shared memory segment using:\nID: %d\tSize:%d\n",memid,memsize);
          ::mf::LogError("shmFail")<<"failed to attach or create a shared memory segment using ID: "<<memid<<" Size: "<<memsize;
	  exit(FAIL);
	} // endif shmid
      } // endif shmid > 0      
            
      shmptr_start   = shmptr;
      shmptr_int     = (int*) shmptr;
      shmptr_end     = shmptr_start + memsize;
      shmptr_current = shmptr_start;
    } // endif memid != 0x0
  } // endif memid != -1


  DataMap   = (SHMSpillDataMap*)shmptr_start;
  AccelData = (AccelEventData*) (DataMap + 2);

  (*DataMap).data_start   = (uint32_t*) AccelData;
  (*DataMap).data_end     = (uint32_t*) shmptr_end;
  (*DataMap).data_insert  = AccelData;
  (*DataMap).data_entries = 0;


  /************************************************************/

  std::cout << "Sending spills..." << std::endl;
  ::mf::LogInfo("Spills")<<"Sending spills...";

  bool events_in_buffer = false;

  int bytesRead=0;

  bool timeSync = false;
  int syncCount = 0;

  // Main event read loop
  // We stay in this loop until told to exit it
  while (read_events) {


    sigsetjmp(termbuffer,1);                                  // Set our re-entry point for signal handling

    // If we are returning from a handler, then we check to see if we have
    if(termination_pending){
      // Check to see if we have events still in the buffer      
      if(events_in_buffer == false){
	// We have no events to process so we just exit out cleanly
	exit(0);
      }
      // We must have some events in the buffer still so we go though and process them
      // but we skip reading the device (i.e. we don't accumulate more events)
      // bytesRead and buffer should have the right values and addresses still
    }else{
      // We read the events from the FPGA queue
      bytesRead = mtu.readEvents((uint8_t *)buffer, 4096);
      if(bytesRead > 0){ events_in_buffer = true;}            // As soon as we have events in the buffer we note it.
    }

    //printf ("bytes read = %d\n", bytesRead);
    if (bytesRead%10 != 0) {
      std::cout << "I expected to read an integer multiple of 10 bytes.\n"
                << "I got " << bytesRead << " bytes instead.\n"
                << "This is weird. Alert an expert!!"<<std::endl;
      ::mf::LogError("Bytes")<<"I expected to read an integer multiple of 10 bytes. "
                             <<" I got "<<bytesRead<<" bytes instead. "
                             <<" This is weird. Alert an expert!!";
    }

    uint64_t latest1D = 0x0; //keep track of 1Ds for BNBtclk

    uint64_t latestTCR = 0x0; // Keep track of the TCR events
    uint64_t oneSecondBoundary = 0;
    uint64_t oneSecondBoundary2 = 0;
    int64_t deltaT = 0;
    int64_t deltaT2 = 0;
    const uint64_t ONE_SECOND = 64000000;


    for (int i=0; i<bytesRead/10; ++i)
    {
      bool     send = true;
      uint16_t evt = buffer[i*5+0];
      uint64_t novatime = 0;
      for (int j=0; j!=4; ++j) {
        novatime += ((uint64_t)buffer[i*5+(j+1)])<<(16*j);
      } // endfor j


      if(BaseClock128){
	// If the base clock of the TDU is 128 MHz, we down 
	// convert the timestamp of the event to 64 MHz for 
	// computation of the NOvA time

	// Down shift 1 bit to go from 128MHz to 64MHz time base
	novatime = novatime >> 1;
      } // endif BaseClock128

      // Subtract 15 leap seconds if running with an older firmware
      if (subfifteen) novatime -= fifteensecs;

      // decode time into spillInfoXml
      NssUtil::gpstimeToString(novatime, str);
      spillInfoXml[0] = str;

      //decode evt into spillInfoXml
      NssSpillInfo::getSpillTypeFromEvent(spillType,evt);
      spillInfoXml[1] = spillType;

      int theAccelEvt;

      // Booster Spill Trigger Conditions
      // These are the coincidence between the $1D and $1F
      if (tappCfg.BNBtclkuse1Dand1F()) {

	theAccelEvt = evt & 0xFF;

	switch( theAccelEvt){
	case 0x1D:
	  // When a $1D is received a trigger is NOT issues immediately
	  // instead the time of the latest $1D is cached/updated to
	  // allow us to form a delayed coincidence with the $1F
          send     = false;
          latest1D = novatime;
          //std::cout << "Not sent: 0x" << std::hex << evt << " at 0x"
          //          << std::hex << novatime << std::dec << std::endl;
	  break;
	case 0x1F:
	  // When a $1F is received we look to see if it is within the
	  // specified coincidence window with the last $1D
          if ( (novatime < (latest1D + tappCfg.BNBtclk1Ddelay())) ||
	       (novatime > (latest1D + tappCfg.BNBtclk1Ddelay()
			    + tappCfg.BNBtclk1Dwidth()))
	      )
          {
	    // If the $1F does not fall within the coincidence window
	    // then we do NOT send the trigger out

	    // std::cout << "Not sent: 0x" << std::hex << evt << " at 0x"
            //         << std::hex << novatime << std::dec << std::endl;
            send = false;
          }
	  break;
	default:
	  send = true;
	  break;
	} // endswitch

      }

      // Check for the TCR Reference Signals
      if (evt == 0x041B){
	// We don't issue triggers on the references, instead we check thier offset 
	// from the canonical 1 second GPS boundary

	latestTCR = novatime;

	oneSecondBoundary  = novatime - (novatime % ONE_SECOND);   // Compute the 1 second boundary
        oneSecondBoundary2 = novatime - (novatime % ONE_SECOND) + ONE_SECOND; //nearest one second boundary could be in the future, check for that
	deltaT             = llabs(latestTCR - oneSecondBoundary);        // Compute the time delta
        deltaT2            = llabs(latestTCR - oneSecondBoundary2); //Compute the delta for the next second

        if (deltaT2 < deltaT) deltaT = deltaT2; //take the smallest delta

	// Check if the time delta is within some threshold of the 1 second boundary
	if(llabs(deltaT) > (int64_t)tappCfg.TCRErrorThreshold()){
	  // If it is not then issue errors to the system
	  ::mf::LogError("TimeDrift") << "TDU Drift ERROR: " << hostname 
				   << " Time Delta from Reference: " << deltaT 
				   << " at Novatime: " << novatime 
				   << std::endl;  
	}else{
	  ::mf::LogInfo("TimeDrift") << "TDU drift Nominal: " << hostname 
				   << " Time Delta from Reference: " << deltaT 
				   << " at Novatime: " << novatime 
				   << std::endl;  
	  
	}

        //wait 300 TCR triggers (5 minutes) before checking for jumps in the trigger.  This is to get over the initial backlog of triggers in the buffer when the server starts
        if (tcrCount < 300) {
          tcrCount++;
        }
        else {
          deltaT = latestTCR - lastTCR_time - ONE_SECOND; //compare the difference between the current and previous TCR trigger to a 1 second interval
	  if(llabs(deltaT) > (int64_t)tappCfg.TCRErrorThreshold()){
	    // If it is not then issue errors to the system
	    ::mf::LogError("TimeSync") << "TDU Sync ERROR: " << hostname 
				     << " Current TCR trigger: " << latestTCR
				     << " previous trigger: " << lastTCR_time
                                     << " differ by: " << (int64_t)latestTCR - (int64_t)lastTCR_time
				     << std::endl;  
	  }else{
	    ::mf::LogInfo("TimeSync") << "TDU Sync Nominal: " << hostname 
				     << " Current TCR trigger: " << latestTCR
				     << " previous trigger: " << lastTCR_time
                                     << " differ by: " << (int64_t)latestTCR - (int64_t)lastTCR_time
				     << std::endl;  
	  
	  }
        }


	//send = false;
        lastTCR_time = novatime;
      }

      // If we have determined that we want to send this
      // event out then we proceed with this block, else
      // we loop back and look for more events

      if(send){
	// First copy the data out into the shared memory segment for logging/diagnostics
     
	// Load the information into a structure
	theAccelData.evt_type = spillType;
	theAccelData.time     = novatime;
	theAccelData.evt_no++;

	// Now do a single copy into the shared memory area
	memcpy((*DataMap).data_insert,&theAccelData,sizeof(theAccelData));

#if 0
	printf("SHM Start: 0x%08x\tDM Insert: 0x%08x\tAccelDat: 0x%08x\tSize: %u\n",
	       shmptr_start,
	       (*DataMap).data_insert,
	       &theAccelData,
	       sizeof(theAccelData));
#endif
		
	// Update the insertion points to point to the next area or back to the
	// start of the data buffer
	(*DataMap).data_insert++;
  	if((uint32_t*)(*DataMap).data_insert > (*DataMap).data_end){
  	  (*DataMap).data_insert = (struct AccelEventData*)(*DataMap).data_start;
	}

  	(*DataMap).data_entries++;
	lastspill_time = novatime;
#if 0
	// Print out the resulting trigger time to standard out
	printf("%d, %lld, %016llx, %10lld, %016llx, %10lld, %016llx\n", spillType,novatime, novatime,
	novatime & 0x1FFFFFFFF,
	novatime & 0x1FFFFFFFF, 
	novatime - lastspill_time,
	novatime - lastspill_time);
#endif
      }

      if (send) {
        if (c.execute("Spill", spillInfoXml, result)) {
          //std::cout << "Spill method result: " << result << "\n";
          //std::cout << "Type " 
          //          << NssSpillInfo::getSpillNameFromType(spillType)
          //          << "(evt 0x" << std::hex << evt << ") at 0x"
          //          << std::hex << novatime << std::dec << std::endl;
          ::mf::LogInfo("SpillSent")<<"Spilltype: "
                                    <<NssSpillInfo::getSpillNameFromType(spillType)
                                    <<"("<<spillType<<")"
                                    <<" (evt 0x"<<std::hex<<evt<<") novaTime 0x"
                                    <<std::hex<<novatime<<std::dec<<" "<<novatime;

          //check that the time stamp for the event is not 0x0.  This indicates that the TDU
          //is most likely missing a time sync.  Correct this by sending a sync to the TDU.
          if (novatime == 0){
            //let spills pass before issuing a sync again.
            if (timeSync == true){
              syncCount++;
              if (syncCount < 100) continue;
            }
            std::cout<<"!!!The time stamp for this event was 0!!! Attempting sync"<<std::endl;
            //check that the status register is ok to sync
            uint32_t status = mtu.getRegister(cnf.statusReg);
            if ((status & 0x8008) && !(status & 0x8000)){
              std::cout<<"It is ok to issue a sync"<<std::endl;
              ctrl = 0x0;
              mtu.getRegister(cnf.ctrlReg);
              ctrl |= 0x0420;
              mtu.setRegister(cnf.ctrlReg, ctrl);
              usleep(4000000);//sleep 4 seconds to allow sync
              timeSync = true;
              syncCount = 0;
            }
            else {
              std::cout<<"NssTDUApp: Error trying to sync TDU.  Time stamps are not expected to be correct.  Exiting server."<<std::endl;
              ::mf::LogError("Sync")<<"NssTDUApp: Error trying to sync detector, time stamps are likely incorrect.  Aborting spill server.";
              return false;
            }
          }
        }
        else {
          std::cout << "NssTDUApp: Error sending 'Spill'\n";
          ::mf::LogError("Spill")<<"NssTDUApp: Error sending 'Spill'";
        }
      } // endif send
    } // endfor i  (accel event unpack loop)
    events_in_buffer = false;

    // If there is a termination signal pending then when 
    // we reach this point we are going to exit instead of 
    // spinning back around to read the next events
    if(termination_pending){
      exit(0);
    }

    // Sleep before reading the TDU event queue
    usleep(SPILL_POLL_DELAY);

  } // endwhile
 
  return FALSE;
}
