#include <stdio.h>
#include "NovaSuperNova/SNDataBuffer.h"

using namespace supernova;

/**
* @brief Construct empty DataBuffer
* Given parameters for DataBlocks, which will be created, 
* when you try to fill a buffer with timestamp out of range,
* or when explicitly extend buffer by
* addBlockFront(), addBlockBack(), extendBuffer()
*
* @param blockDur: Duration in time of DataBlocks to be created 
* @param blockLen: Length of DataBlocks to be created
*/
DataBuffer::DataBuffer(const TimeType & blockDur, size_t blockLen):
	VirtualDataBlock(TimeRange(0,0)),_blockDuration(blockDur),_blockLength(blockLen)
{
}

DataBuffer::~DataBuffer()
{
  _blocks.clear();
}

/**
* @brief Fill given pair of values to buffer
*
* @param t - timestamp, to find the bin, where to fill the d value
* @param d - value to fill
* If buffer doesn't contain given timestamp, it will be extended by adding DataBlocks
* until given timestamp is contained.
*/
void DataBuffer::fill(const TimeType& t, const DataType& d)
{ 
   if(!contains(t))extendBuffer(t);
   DataBlock& block=findBlock(t);
   //fill given value to block
   block.fill(t,d);
}

/**
* @brief Update starting end ending time of this buffer, using information from blocks
*/
void DataBuffer::updateTimeLimits()
{
   if(empty()){
        _time0=_time1; 
        printf("Block is EMPTY => now limits are %ld-%ld\n",_time0,_time1);
        return;
    }
   _time0=_blocks.back().startTime();
   _time1=_blocks.front().endTime();
}

/**
* @brief Add default block (with duration=_blockDuration,length=_blockLength) to back
* @throws supernova::errors::basic_error if current starting time is 0
*/
void DataBuffer::addBlockBack()
{
  if(_time0<_blockDuration) 
    throw errors::basic_error("Cannot add previous DataBlock, starting time is 0");
  pushBlockBack(DataBlock(_blockLength,TimeRange(_time0-_blockDuration,_time0)));
}

/**
* @brief Add default block (with duration=_blockDuration,length=_blockLength) to front
*/
void DataBuffer::addBlockFront()
{
   pushBlockFront(DataBlock(_blockLength,TimeRange(_time1,_time1+_blockDuration)));
}

/**
* @brief Extend this buffer by adding standard DataBlocks(duration=_blockDuration,length=_blockLength)
* to cover given timestamp.
*
* @param t - timestamp, which we want to cover with buffer
* if is in the past (t> timeStart()) - extend in the beginning of the blocks queue, 
* if is in the future (t> timeEnd()) - extend in the end of the queue
*/
void DataBuffer::extendBuffer(const TimeType& t)
{
  if(empty()){
    if(_time1==0){
        //this is the first message received ever
        _time0=t;
        _time1=t;
        }
    else{
        //we need to find block start, based on last known block end
        int NblocksSkip=(t-_time0)/_blockDuration;
        printf("Skipping %d blocks!\n",NblocksSkip);
        _time1=_time0+_blockDuration*NblocksSkip;
     }
  }
  bool extendInFuture=(t>=endTime());
  while(!contains(t)){
     (extendInFuture)?addBlockFront():addBlockBack();
  }
}

/**
* @brief Access data at n-th bin
* Note, that this procedure is not very fast (cycles through all blocks),
* so one should use iterators when possible.
* @see begin(), end()
*
* @param n - bin position, counting from beginning of first block
*
* @return reference to data in this bin
*/
DataType& DataBuffer::dataAtPos(size_t n)
{
   std::deque<DataBlock>::iterator it;
   for(it=_blocks.begin();it!=_blocks.end();++it){
     if(n<it->size())break;
     n-=it->size(); continue;
   } 
   return it->dataAtPos(n); 
}

/**
* @brief find DataBlock, containing given timestamp
* @throws errors::out_of_range in case given timestamp is not inside this buffer range
* @throws errors::basic_error if DataBlock is not found
* @return reference to the found DataBlock
*/
DataBlock& DataBuffer::findBlock(const TimeType& t)
{
  if(!contains(t))throw errors::out_of_range(t,startTime(),endTime());
  //start searching backwards, since mostly probable 
  std::deque<DataBlock>::reverse_iterator rit;
  for(rit=_blocks.rbegin();rit!=_blocks.rend();++rit){
     if( rit->contains(t))return *rit;
  }
  throw errors::basic_error("Data block not found");
}

